# -*- coding: utf-8 -*-
"""
##############################################################################

File containing a class intend to facilitate machine learning with peptides or other.
The functions are based on the package scikit learn.
Model available: 'svm', 'linear_svm', 'knn', 'sgd', 'lr','rf', 'gnb', 'nn','gboosting'

Authors: Ana Marta Sequeira

Date:06/2019 altered 01/2021

Email:

##############################################################################
"""

import pandas as pd
import time
import seaborn as sns
import warnings
import numpy as np
import matplotlib.pyplot as plt
from sklearn.ensemble import RandomForestClassifier
from numpy import interp
from itertools import cycle
from sklearn.metrics import *
from sklearn.pipeline import Pipeline
from sklearn.model_selection import cross_validate
from sklearn.svm import SVC, LinearSVC
from sklearn.neighbors import KNeighborsClassifier
from sklearn.linear_model import SGDClassifier, LogisticRegressionCV, LogisticRegression
from sklearn.naive_bayes import GaussianNB
from sklearn.neural_network import MLPClassifier
from sklearn.ensemble import GradientBoostingClassifier
from sklearn.svm import SVR, LinearSVR

from sklearn.linear_model import LogisticRegression
from sklearn.linear_model import TweedieRegressor
from sklearn.neighbors import KNeighborsRegressor
from sklearn.ensemble import RandomForestRegressor, GradientBoostingRegressor, AdaBoostRegressor
from sklearn.experimental import enable_hist_gradient_boosting
from sklearn.ensemble import HistGradientBoostingRegressor
from sklearn.gaussian_process import GaussianProcessRegressor

from sklearn.linear_model import SGDRegressor
from sklearn.model_selection import learning_curve
from sklearn.model_selection import ShuffleSplit, train_test_split
from sklearn.metrics import classification_report
from sklearn.model_selection import validation_curve
from sklearn.metrics import make_scorer
from sklearn.metrics import matthews_corrcoef
from sklearn.metrics import roc_curve, auc
from sklearn.preprocessing import label_binarize
from sklearn.multiclass import OneVsRestClassifier
from sklearn.metrics import roc_auc_score
from sklearn.metrics import explained_variance_score, max_error, mean_absolute_error, mean_squared_error, \
    mean_squared_log_error, median_absolute_error, r2_score, mean_poisson_deviance, mean_gamma_deviance
from .utils import timer
from .parameters_shallow import param_shallow
from .param_optimizer import ParamOptimizer

warnings.filterwarnings("ignore")
# Seed for randomization. Set to some definite integer for debugging and set to None for production
seed = None


class ShallowML:
    """
    The MachineLearning class aims to process different models of machine learning with peptides.
    Model available: 'svm', 'linear_svm', 'knn', 'sgd', 'lr','rf', 'gnb', 'nn','gboosting'
    Based on scikit learn.
    """

    # todo acrescentei problem_type
    def __init__(self, x_train, x_test, y_train, y_test, report_name=None, columns_names=None,
                 problem_type='src', classifier = None):
        """
        init function. When the class is called a dataset containing the features values and a target column must
        be provided.
        :param x_train: dataset with features or encodings for training
        :param x_test: dataset with features or encodings for evaluation
        :param y_train: class labels for training
        :param y_test: class labels for testing
        :param report_name: str. If not none it will generate a report txt with the name given) with results by functions
        called within class. Default None
        :param columns_names: Names of columns. important if features importance want to be analysed. None by default.
        """
        # self.X_data = pd.DataFrame(X_data) # take X_data off
        # self.Y_data = Y_data
        self.x_train = x_train
        self.x_test = x_test
        # self.x_dval = X_dval
        self.y_train = y_train
        self.y_test = y_test
        # self.y_dval = y_dval.ravel()
        # y = column_or_1d(y, warn=True)

        # ver se vale a pena aceitar de fora
        self.model_name = None
        self.feat_table = None
        self.classifier = classifier

        if columns_names is not None:
            self.columns = columns_names
        else:
            try:
                self.columns = self.x_train.columns
            except Exception as e:
                print(e)
                print('no features names listed')

        self.problem_type = problem_type
        if self.problem_type == 'src':
            self.main_score = make_scorer(matthews_corrcoef)
            self.final_units = len(np.unique(y_train))
        elif self.problem_type == 'regression':
            self.main_score = 'neg_root_mean_squared_error'
        self.report_name = report_name
        if self.report_name:
            self._report(str(self.report_name))

    def _report(self, info, dataframe=False, **kwargs):
        filename = str(self.report_name)
        with open(filename, 'a+') as file:
            if dataframe is True:
                info.to_csv(file, sep='\t', mode='a', **kwargs)
            elif isinstance(info, str):
                file.writelines(info)
            else:
                for l in info:
                    file.writelines('\n{}'.format(l))
    def get_model(self):
        return self.classifier

    def train_best_model(self, model, scaler=None,
                         score=None,
                         cv=10, optType='gridSearch', param_grid=None,
                         n_jobs=5,
                         random_state=1, n_iter=15, refit=True, **params):
        # todo put the name here oh the siglas. lr - logistic regression ....

        """
        This function performs a parameter grid search or randomizedsearch on a selected classifier model and training data set.
        It returns a scikit-learn pipeline that performs standard scaling (if not None) and contains the best model found by the
        grid search according to the Matthews correlation coefficient or other given metric.
        :param model_name: {str} model to train. Choose between 'svm', 'linear_svm', 'knn', 'sgd', 'lr','rf', 'gnb', 'nn','gboosting'
        :param model: scikit learn model
        :param scaler: {scaler} scaler to use in the pipe to scale data prior to training (integrated  in pipeline)
         Choose from
            ``sklearn.preprocessing``, e.g. 'StandardScaler()', 'MinMaxScaler()', 'Normalizer()' or None.None by default.
        :param score: {metrics instance} scoring function built from make_scorer() or a predefined value in string form
            (choose from the scikit-learn`scoring-parameters <http://scikit-learn.org/stable/modules/model_evaluation.html#scoring-parameter>`_).
        :param optType: type of hyperparameter optimization to do . 'gridSearch' or 'randomizedSearch'
        :param param_grid: {dict} parameter grid for the grid or randomized search
        (see`sklearn.grid_search <http://scikit-learn.org/stable/modules/model_evaluation.html>`_).
        :param cv: {int} number of folds for cross-validation.
        :param n_iter: number of iterations when using randomizedSearch optimizaton. default 15.
        :param n_jobs: {int} number of parallel jobs to use for calculation. if ``-1``, all available cores are used.
        10 by default.
        :param refit: wether to refit the model accordingly to the best model . Default True.
        :param random_state: random_state 1 by default.
        :param **params: params to integrate in scikit models
        :return: best classifier fitted to training data.
        """
        start = time.perf_counter()
        print("performing {}...".format(optType))

        saved_args = locals()
        if model is str:
            model = model.lower()
            self.model_name = model_name

        # todo change the SVM to SVC   linear_svc
        if model == 'svc':
            pipe_model = Pipeline([('scl', scaler),
                                   ('clf', SVC(random_state=random_state, **params))])

        elif model == 'linear_svc':  # scale better to large number of samples
            pipe_model = Pipeline([('scl', scaler),
                                   ('clf', LinearSVC(random_state=random_state, **params))])

        elif model == 'rf':
            pipe_model = Pipeline([('scl', scaler),
                                   ('clf', RandomForestClassifier(random_state=random_state, **params))])

        elif model == 'gboosting':
            pipe_model = Pipeline([('scl', scaler),
                                   ('clf', GradientBoostingClassifier(random_state=random_state, **params))])

        elif model == 'knn':
            pipe_model = Pipeline([('scl', scaler),
                                   ('clf', KNeighborsClassifier(**params))])

        elif model == 'sgd':
            pipe_model = Pipeline([('scl', scaler),
                                   ('clf', SGDClassifier(random_state=random_state, **params))])

        elif model == 'lr':
            pipe_model = Pipeline([('scl', scaler),
                                   ('clf', LogisticRegression(random_state=random_state, **params))])

        elif model == 'gnb':
            pipe_model = Pipeline([('scl', scaler),
                                   ('clf', GaussianNB(**params))])

        elif model == 'nn':
            pipe_model = Pipeline([('scl', scaler),
                                   ('clf', MLPClassifier(**params))])

        # regression models
        elif model == 'svr':
            pipe_model = Pipeline([('scl', scaler),
                                   ('clf', SVR(**params))])

        elif model == 'linear_svr':
            pipe_model = Pipeline([('scl', scaler),
                                   ('clf', LinearSVR(random_state=random_state, **params))])

        elif model == 'rfr':
            pipe_model = Pipeline([('scl', scaler),
                                   ('clf', RandomForestRegressor(random_state=random_state, **params))])

        elif model == 'gboostingr':
            pipe_model = Pipeline([('scl', scaler),
                                   ('clf', GradientBoostingRegressor(random_state=random_state, **params))])

        elif model == 'histgboostingr':
            pipe_model = Pipeline([('scl', scaler),
                                   ('clf', HistGradientBoostingRegressor(random_state=random_state, **params))])

        elif model == 'adaboostr':
            pipe_model = Pipeline([('scl', scaler),
                                   ('clf', AdaBoostRegressor(random_state=random_state, **params))])

        elif model == 'tweedier':
            pipe_model = Pipeline([('scl', scaler),
                                   ('clf', TweedieRegressor(**params))])

        elif model =='knr':
            pipe_model = Pipeline([('scl', scaler),
                                   ('clf', KNeighborsRegressor(**params))])

        elif model == 'gpr':
            pipe_model = Pipeline([('scl', scaler),
                                   ('clf', GaussianProcessRegressor(**params))])

        elif model == 'sgdr':
            pipe_model = Pipeline([('scl', scaler),
                                   ('clf', SGDRegressor(random_state=random_state,**params))])
        # todo add light GBM ? alterar nomes do Gboosting para GBc e GBr ? add XGboost? adicionar para classificacao tb

        # todo alter
        elif model is str:
            # keras classifier
            print("Model not supported, please choose between "
                  "src models: 'svc', 'knn', 'sgd', 'rf', 'gnb', 'nn', 'gboosting'"
                  "or regression models:  ")
            return

        else:
            # todo confirm if is working
            pipe_model = Pipeline([('scl', scaler),
                                   ('clf', model)])

        # define scores
        # todo change this. TEST it. make a self.score ?
        # if model in ['svc', 'linear_svc', 'rf', 'gboosting', 'knn', 'sgd', 'lr', 'gnb', 'nn'] and score is None:
        #     score = make_scorer(matthews_corrcoef)
        # elif model in ['svr', 'linear_svr']:
        #     score = r2_score
        # if score is None:  # did not change with models (because is a different one) and did not specify score
        #     print("Please specify a metric score (available on scikit learn)"
        #           "for example accuraccy or make_scorer(matthews_corrcoef) for src and r2_score for "
        #           "regression problems. Check https://scikit-learn.org/stable/modules/classes.html#module-sklearn.metrics"
        #           "for more info")
        #     return

        if score is None:
            score = self.main_score  # accordingly with src or regression
        # or ifself.classification_type == ... score ==
        # if self.src is None: parte dos model in

        # retrieve default grids
        if param_grid is None:
            param = param_shallow()
            if optType == 'gridSearch':
                param_grid = param[model]['param_grid']
            else:
                param_grid = param[model]['distribution']

        po = ParamOptimizer(optType=optType, estimator=pipe_model, paramDic=param_grid,
                            dataX=self.x_train, datay=self.y_train,
                            cv=cv, n_iter_search=n_iter, n_jobs=n_jobs, scoring=score, model_name=self.model_name,
                            refit=refit)
        gs = po.get_opt_params()

        # # summaries
        list_to_write_top_3_models = po.report_top_models(gs)

        s1, s2, s3, df = po.report_all_models(gs)  # metrics for the best model. dataframe with all
        # except:
        #     print('not print all models')
        # Set the best parameters to the best estimator
        # todo take out of this function and put otside. a function to calculate a single model. put self.best calssfier and a function to get it
        best_classifier = gs.best_estimator_
        best_classifier_fit = best_classifier.fit(self.x_train, self.y_train)
        self.classifier = best_classifier_fit
        final = time.perf_counter()
        run_time = final - start

        # write report
        if self.report_name is not None:
            self._report(['===TRAIN MODELS===\n', self.train_best_model.__name__, saved_args])
            self._report(list_to_write_top_3_models)
            self._report([s1, s2, s3, f"Finished {self.train_best_model.__name__} in {run_time:.4f} secs\n\n"])
            self._report(df, dataframe=True, float_format='%.3f')

        return best_classifier_fit

    def cross_val_score_model(self, model,
                              score=None,
                              cv=10,
                              n_jobs=10,
                              random_state=1, **params):

        """
        This function performs cross validations core on a selected classifier model and training data set.
        It returns the scores across the different folds, means and standard deviations of these scores.
        :param model_name: {str} model to train. Choose between 'svm', 'knn', 'sgd', 'lr','rf', 'gnb', 'nn','gboosting'
        :param model: scikit learn machine learning model object
        :param score: {metrics instance} scoring function built from make_scorer() or a predefined value in string form
            (choose from the scikit-learn`scoring-parameters <http://scikit-learn.org/stable/modules/model_evaluation.html#scoring-parameter>`_).
        :param cv: {int} number of folds for cross-validation score.
        :param n_jobs: {int} number of parallel jobs to use for calculation. if ``-1``, all available cores are used.
        10 by default.
        :param random_state: random_state 1 by default.
        :param **params: params to integrate in scikit models
        :return: dataframe with scores per fold, mean and std of these scores.
        """
        start = time.perf_counter()
        print("performing cross val score with {} folds".format(cv))

        saved_args = locals()

        if score is None:
            score = self.main_score
        if model == 'svc':
            clf = SVC(random_state=random_state, **params)

        elif model == 'linear_svc':  # scale better to large number of samples
            clf = LinearSVC(random_state=random_state, **params)

        elif model == 'rf':
            clf = RandomForestClassifier(random_state=random_state, **params)

        elif model == 'gboosting':
            clf = GradientBoostingClassifier(random_state=random_state, **params)

        elif model == 'knn':
            clf = KNeighborsClassifier(**params)

        elif model == 'sgd':
            clf = SGDClassifier(random_state=random_state, **params)

        elif model == 'lr':
            clf = LogisticRegression(random_state=random_state, **params)

        elif model == 'gnb':
            clf = GaussianNB(**params)

        elif model == 'nn':
            clf = MLPClassifier(**params)

        # todo add regression models
        elif model == 'gpr':
           clf = GaussianProcessRegressor(**params)

        elif model is str:
            print("Model not supported, please choose between 'svc', 'knn', 'sgd', 'rf', 'gnb', 'nn', 'gboosting', 'gpr' ")
            return
        else:
            clf = model
        # retrieve default grids
        scores = cross_validate(clf, self.x_train, self.y_train, cv=cv, scoring=score)
        # print(scores)
        scores = pd.DataFrame(scores)
        scores.loc['mean'] = scores.mean()
        scores.loc['std'] = scores.std()

        # try:
        #     scores['test_mean_score'] = np.mean(scores['test_score'])
        #     scores['test_std_score'] = np.std(scores['test_score'])
        #
        # except:
        #     for scorer in score:
        #         scores['test_mean_%s' % (scorer)] = np.mean(scores['test_%s' % (scorer)])
        #         scores['test_std_%s' % (scorer)] = np.std(scores['test_%s' % (scorer)])

        final = time.perf_counter()
        run_time = final - start

        # write report
        if self.report_name is not None:
            self._report(['===TRAIN MODELS===\n', self.cross_val_score_model.__name__, saved_args])
            self._report(scores)
            self._report(f"Finished {self.cross_val_score_model.__name__} in {run_time:.4f} secs\n\n")

        return scores

    ####################################################################################################################
    # EVALUATE
    ####################################################################################################################
    # todo join this to use with deep
    def conf_matrix_seaborn_table(self, conf_matrix=None, classifier=None, path_save='', show=True,
                                  square=True, annot=True, fmt='d', cbar=False, **params):
        plt.clf()
        if conf_matrix is None:
            y_pred = classifier.predict(self.x_test)
            mat = confusion_matrix(self.y_test, y_pred)
        else:
            mat = conf_matrix

        sns.heatmap(mat.T, square=square, annot=annot, fmt=fmt, cbar=cbar, **params)
        plt.xlabel('true label')
        plt.ylabel('predicted label')
        if path_save is not None:
            plt.savefig(fname=path_save)
        if show is True:
            plt.show()
        plt.clf()

    def score_testset_classification(self, classifier=None):
        """
        Returns the tests set scores for the specified scoring metrics in a ``pandas.DataFrame``. The calculated metrics
        are Matthews correlation coefficient, accuracy, precision, recall, f1 and area under the Receiver-Operator Curve
        (roc_auc). See `sklearn.metrics <http://scikit-learn.org/stable/modules/classes.html#sklearn-metrics-metrics>`_
        for more information.
        :param classifier: {classifier instance} pre-trained classifier used for predictions. If None, will use the
        one trained in the class.
        :return: ``pandas.DataFrame`` containing the cross validation scores for the specified metrics.
        """
        if classifier is None:  # priority for the classifier in function
            classifier = self.classifier
        saved_args = locals()
        cm2 = []
        scores = {}
        y_pred = classifier.predict(self.x_test)
        try:
            y_prob = classifier.predict_proba(self.x_test)
        except:
            y_prob = None
        scores['Accuracy'] = accuracy_score(self.y_test, y_pred)
        scores['MCC'] = matthews_corrcoef(self.y_test, y_pred)
        if y_prob is not None:
            scores['log_loss'] = log_loss(self.y_test, y_prob)

        if self.final_units > 2:
            # multiclass
            scores['f1 score weighted'] = f1_score(self.y_test, y_pred, average='weighted')
            scores['f1 score macro'] = f1_score(self.y_test, y_pred, average='macro')
            scores['f1 score micro'] = f1_score(self.y_test, y_pred, average='micro')
            if y_prob is not None:
                scores['roc_auc ovr'] = roc_auc_score(self.y_test, y_prob, average='weighted', multi_class='ovr')
                y_test = self.y_test
                # y_test = y_test.reshape(y_test.shape[0])  # roc auc ovo was giving error
                scores['roc_auc ovo'] = roc_auc_score(y_test, y_prob, average='weighted', multi_class='ovo')
            scores['precision'] = precision_score(self.y_test, y_pred, average='weighted')
            scores['recall'] = recall_score(self.y_test, y_pred, average='weighted')
            cm2 = multilabel_confusion_matrix(self.y_test, y_pred)

        else:
            # binary
            scores['f1 score'] = f1_score(self.y_test, y_pred)
            scores['roc_auc'] = roc_auc_score(self.y_test, y_pred)
            precision, recall, thresholds = precision_recall_curve(self.y_test, y_pred)
            scores['Precision'] = precision
            scores['Recall'] = recall
            tn, fp, fn, tp = confusion_matrix(self.y_test, y_pred).ravel()
            scores['fdr'] = float(fp) / (tp + fp)
            scores['sn'] = float(tp) / (tp + fn)
            scores['sp'] = float(tn) / (tn + fp)

        report = classification_report(self.y_test, y_pred, output_dict=False)
        cm = confusion_matrix(self.y_test, y_pred)
        cm2 = None

        if self.report_name is not None:
            self._report(['===SCORING TEST SET ===\n', self.score_testset.__name__, saved_args, 'report\n', report,
                          '\nconfusion_matrix\n', cm, '\nmultilabel confusion matrix\n', cm2, '\nscores report\n'])
            scores_df = pd.DataFrame(scores.keys(), columns=['metrics'])
            scores_df['scores'] = scores.values()
            self._report(scores_df, dataframe=True, float_format='%.4f', index=False)

            self.conf_matrix_seaborn_table(conf_matrix=cm,
                                           path_save=str(self.report_name[-4:] + 'confusion_matrix.png'),
                                           show=False)

        return scores, report, cm, cm2
    # todo add this fucntion?
    # see if it is worth it. put it with not priniting . giving something closer to the src report scikit learn
    def regression_report(self,y_true, y_pred):

        error = y_true - y_pred
        percentil = [5, 25, 50, 75, 95]
        percentil_value = np.percentile(error, percentil)

        metrics = [
            ('mean absolute error', mean_absolute_error(y_true, y_pred)),
            ('median absolute error', median_absolute_error(y_true, y_pred)),
            ('mean squared error', mean_squared_error(y_true, y_pred)),
            ('max error', max_error(y_true, y_pred)),
            ('r2 score', r2_score(y_true, y_pred)),
            ('explained variance score', explained_variance_score(y_true, y_pred))
        ]

        print('Metrics for regression:')
        for metric_name, metric_value in metrics:
            print(f'{metric_name:>25s}: {metric_value: >20.3f}')

        print('\nPercentiles:')
        for p, pv in zip(percentil, percentil_value):
            print(f'{p: 25d}: {pv:>20.3f}')
        return None

    # todo add this function !!!!
    # https://medium.com/analytics-vidhya/mae-mse-rmse-coefficient-of-determination-adjusted-r-squared-which-metric-is-better-cd0326a5697e
    def score_testset_regression(self, classifier=None):
        """
        Returns the tests set scores for the specified scoring metrics in a ``pandas.DataFrame``. The calculated metrics
        are Matthews correlation coefficient, accuracy, precision, recall, f1 and area under the Receiver-Operator Curve
        (roc_auc). See `sklearn.metrics <http://scikit-learn.org/stable/modules/classes.html#sklearn-metrics-metrics>`_
        for more information.
        :param classifier: {classifier instance} pre-trained classifier used for predictions. If None, will use the
        one trained in the class.
        :return: ``pandas.DataFrame`` containing the cross validation scores for the specified metrics.
        """
        if classifier is None:  # priority for the classifier in function
            classifier = self.classifier
        saved_args = locals()
        cm2 = []
        scores = {}
        y_pred = classifier.predict(self.x_test)
        # todo put abs in negative values? they give negative values so the model try to get the lower score rather than the higher.
        # https://stackoverflow.com/questions/21443865/scikit-learn-cross-validation-negative-values-with-mean-squared-error
        scores['neg_root_mean_squared_error - RMSE'] = abs(mean_squared_error(self.y_test, y_pred, squared=False))
        scores['neg_mean_squared_error - MSE'] = abs(mean_squared_error(self.y_test, y_pred, squared=True))
        scores['neg_mean_absolute_error - MAE'] = abs(mean_absolute_error(self.y_test, y_pred))
        scores['r2'] = r2_score(self.y_test, y_pred) # It represents the proportion of variance (of y) that has been explained
        # by the independent variables in the model. It provides an indication of goodness of fit and therefore a measure
        # of how well unseen samples are likely to be predicted by the model, through the proportion of explained variance.
        # As such variance is dataset dependent, R² may not be meaningfully comparable across different datasets.
        scores['explained_variance'] = explained_variance_score(self.y_test, y_pred) # The best possible score is 1.0, lower values are worse
        scores['max_error'] = max_error(self.y_test, y_pred) # calculates the maximum residual error. aptures the worst case error between the predicted value and the true value

        try:
            scores['neg_mean_squared_log_error - MSLE'] = abs(mean_squared_log_error(self.y_test, y_pred)) # expected value of the squared logarithmic (quadratic) error or loss.

        except Exception as e:
            print('neg_mean_squared_log_error could not be calculated')
            print(str(e))

        try:
            scores['neg_median_absolute_error'] = abs(median_absolute_error(self.y_test, y_pred)) # robust to outliers. The loss is calculated by taking the median of all absolute differences between the target and the prediction. Do not support multioutput
        except Exception as e:
            print('neg_median_absolute_error could not be calculated')
            print(str(e))

        # todo only for positive values? some of them ? check
        try:
            scores['neg_mean_poisson_deviance'] = abs(mean_poisson_deviance(self.y_test, y_pred))
        except Exception as e:
            print('neg_mean_poisson_deviance could not be calculated')
            print(str(e))
        try:
            scores['neg_mean_gamma_deviance'] = abs(mean_gamma_deviance(self.y_test, y_pred))
        except Exception as e:
            print('neg_mean_gamma_deviance could not be calculated')
            print(str(e))
        try:
            scores['mean_absolute_percentage_error'] = abs(mean_absolute_percentage_error(self.y_test, y_pred)) # sensitive to relative errors. supports multioutput
        except Exception as e:
            print('mean_absolute_percentage_error could not be calculated')
            print(str(e))

        report = self.regression_report(self.y_test, y_pred)
        cm = None
        cm2 = None

        if self.report_name is not None:
            self._report(['===SCORING TEST SET ===\n', self.score_testset.__name__, saved_args, 'report\n', report,
                          '\nscores report\n'])
            scores_df = pd.DataFrame(scores.keys(), columns=['metrics'])
            scores_df['scores'] = scores.values()
            self._report(scores_df, dataframe=True, float_format='%.4f', index=False)

        return scores, report

    # todo add this. Put text explaining return all the same?

    def score_testset(self, classifier=None):
        if self.problem_type == 'src':
            scores, report, cm, cm2 = self.score_testset_classification(classifier=classifier)
            return scores, report, cm, cm2
        elif self.problem_type == 'regression':
            scores, report = self.score_testset_regression(classifier=classifier)
            return scores, report

    @timer
    def plot_roc_curve(self, classifier=None, ylim=(0.0, 1.00), xlim=(0.0, 1.0),
                       title='Receiver operating characteristic (ROC) curve',
                       path_save='plot_roc_curve', show=False):
        """
        Function to plot a ROC curve
        On the y axis, true positive rate and false positive rate on the X axis.
        The top left corner of the plot is the 'ideal' point - a false positive rate of zero, and a true positive rate
        of one, meaning a larger area under the curve (AUC) is usually better.
        :param classifier: {classifier instance} pre-trained classifier used for predictions.
        :param ylim: y-axis limits
        :param xlim: x- axis limits
        :param title: title of plot. 'Receiver operating characteristic (ROC) curve' by default.
        :param path_save: path to save the plot. If None , lot is not saved. 'plot_roc_curve' by default.
        :param show: Whether to display the graphic. False by default.
        :return:
        Needs classifier with probability
        https://scikit-learn.org/stable/auto_examples/model_selection/plot_roc.html
        """
        lw = 2
        if classifier is None:
            classifier = self.classifier
        # binary
        if self.final_units <= 2:
            y_score = classifier.predict(self.x_test)
            # y_score = classifier.predict_proba(X_test)[:,1]
            fpr, tpr, thresholds = roc_curve(self.y_test, y_score)
            roc_auc = auc(fpr, tpr)
            print(roc_auc)
            plt.show()
            plt.figure()
            plt.plot(fpr, tpr, color='darkorange',
                     lw=lw, label='ROC curve (area = %0.2f)' % roc_auc)
            plt.plot([0, 1], [0, 1], color='navy', lw=lw, linestyle='--')
            plt.xlim(xlim)
            plt.ylim(ylim)
            plt.xlabel('False Positive Rate')
            plt.ylabel('True Positive Rate')
            plt.title(title)
            plt.legend(loc="lower right", )


        # multiclass/ multilabel
        elif self.final_units > 2:

            # Binarize the output
            classe = np.unique(self.y_train)
            y_train = label_binarize(self.y_train, classes=classe)
            n_classes = y_train.shape[1]
            y_test = label_binarize(self.y_test, classes=classe)

            # Learn to predict each class against the other
            estimator = OneVsRestClassifier(classifier)
            y_score = estimator.fit(self.x_train, y_train).predict_proba(self.x_test)

            # Compute ROC curve and ROC area for each class
            fpr = dict()
            tpr = dict()
            roc_auc = dict()
            for i in range(n_classes):
                fpr[i], tpr[i], _ = roc_curve(y_test[:, i], y_score[:, i])
                roc_auc[i] = auc(fpr[i], tpr[i])

            # Compute micro-average ROC curve and ROC area
            fpr["micro"], tpr["micro"], _ = roc_curve(y_test.ravel(), y_score.ravel())
            roc_auc["micro"] = auc(fpr["micro"], tpr["micro"])

            # First aggregate all false positive rates
            all_fpr = np.unique(np.concatenate([fpr[i] for i in range(n_classes)]))
            print(all_fpr)
            # Then interpolate all ROC curves at this points
            mean_tpr = np.zeros_like(all_fpr)
            for i in range(n_classes):
                mean_tpr += interp(all_fpr, fpr[i], tpr[i])

            # Finally average it and compute AUC
            mean_tpr /= n_classes

            fpr["macro"] = all_fpr
            tpr["macro"] = mean_tpr
            roc_auc["macro"] = auc(fpr["macro"], tpr["macro"])

            # Plot all ROC curves
            plt.figure()
            plt.plot(fpr["micro"], tpr["micro"],
                     label='micro-average ROC curve (area = {0:0.2f})'
                           ''.format(roc_auc["micro"]),
                     color='deeppink', linestyle=':', linewidth=4)

            plt.plot(fpr["macro"], tpr["macro"],
                     label='macro-average ROC curve (area = {0:0.2f})'
                           ''.format(roc_auc["macro"]),
                     color='navy', linestyle=':', linewidth=4)

            # Get Unique ec
            color_labels = np.unique(self.y_train)

            # List of colors in the color palettes
            rgb_values = sns.color_palette("nipy_spectral", len(color_labels))  # 'set2'
            # Map ec to the colors
            for i, color in zip(range(len(color_labels)), rgb_values):
                plt.plot(fpr[i], tpr[i], color=color, lw=lw,
                         label='ROC curve of class {0} (area = {1:0.2f})'
                               ''.format(i, roc_auc[i]))

            plt.plot([0, 1], [0, 1], 'k--', lw=lw)
            plt.xlim([0.0, 1.0])
            plt.ylim([0.0, 1.05])
            plt.xlabel('False Positive Rate')
            plt.ylabel('True Positive Rate')
            plt.title(title)
            plt.legend(loc="lower right", fontsize='xx-small')
        if path_save is not None:
            plt.savefig(fname=path_save)
        if show is True:
            plt.show()
        plt.clf()

    ####################################################################################################################
    # PLOTS : Model curves
    ####################################################################################################################

    @timer
    def plot_validation_curve(self, param_name, param_range,
                              classifier=None,
                              cv=5,
                              score=None, title="Validation Curve",
                              xlab="parameter range", ylab="MCC", n_jobs=1, show=False,
                              path_save='plot_validation_curve', **params):

        """
        This function plots a cross-validation curve for the specified classifier on all tested parameters given in the
        option 'param_range'.

        :param param_name: {string} parameter to assess in the validation curve plot. For example,
        For SVM,
            "clf__C" (C parameter), "clf__gamma" (gamma parameter).
        For Random Forest,
            "clf__n_estimators" (number of trees),"clf__max_depth" (max num of branches per tree,
            "clf__min_samples_split" (min number of samples required to split an internal tree node),
            "clf__min_samples_leaf" (min number of samples in newly created leaf).
        :param param_range: {list} parameter range for the validation curve.
        :param classifier: {classifier instance} classifier or validation curve (e.g. sklearn.svm.SVC).If None (default)
        uses the classifier inside the class.
        :param cv: {int} number of folds for cross-validation. 5 by default.
        :param score: {metrics instance} scoring function built from make_scorer() or a predefined value in string form
            `sklearn.model_evaluation.scoring-parameter
            <http://scikit-learn.org/stable/modules/model_evaluation.html#scoring-parameter>`_.
        :param title: {str} graph title
        :param xlab: {str} x axis label.
        :param ylab: {str} y axis label.
        :param n_jobs: {int} number of parallel jobs to use for calculation. if ``-1``, all available cores are used.
        1 by default.
        :param path_save: path to save the plot. If None , lot is not saved. 'plot_validation_curve' by default.
        :param show: Whether to display the graphic. False by default.
        **params: other parameters to integrate in the function of validation curve calculus.
        :return: plot of the validation curve.
        """
        if classifier is None:
            classifier = self.classifier
        if score is None:
            score = self.main_score

        train_scores, test_scores = validation_curve(classifier, self.x_train, self.y_train, param_name, param_range,
                                                     cv=cv, scoring=score, n_jobs=n_jobs, **params)

        train_mean = np.mean(train_scores, axis=1)
        train_std = np.std(train_scores, axis=1)
        test_mean = np.mean(test_scores, axis=1)
        test_std = np.std(test_scores, axis=1)

        # plotting
        plt.clf()
        plt.title(title)
        plt.xlabel(xlab)
        plt.ylabel(ylab)
        plt.ylim(0.0, 1.1)
        plt.semilogx(param_range, train_mean, label="Training score", color="b")
        plt.fill_between(param_range, train_mean - train_std, train_mean + train_std, alpha=0.2, color="b")
        plt.semilogx(param_range, test_mean, label="Cross-validation score", color="g")
        plt.fill_between(param_range, test_mean - test_std, test_mean + test_std, alpha=0.2, color="g")
        plt.legend(loc="best")

        if path_save is not None:
            plt.savefig(fname=path_save)
        if show is True:
            plt.show()
        plt.clf()

    @timer
    def plot_learning_curve(self, classifier=None, title='Learning curve', ylim=None,
                            cv=ShuffleSplit(n_splits=10, test_size=0.2, random_state=0),
                            n_jobs=1, train_sizes=np.linspace(.1, 1.0, 5),
                            path_save='plot_learning_curve', show=False, scalability=True, performance=True):
        """
        Plot a learning curve to determine cross validated training and tests scores for different training set sizes.
        IT retrieves graphic representing learning curves, numbers of trainig examples, scores on training sets,
        and scores on tests set and scalability and performance plots if set to True.
        :param classifier: {classifier instance} classifier or validation curve (e.g. sklearn.svm.SVC).If None (default)
        uses the classifier inside the class.
        :param title: title of the plot
        :param ylim: None by default.
        :param cv: cross validation to use. ShuffleSplit(n_splits=10, test_size=0.2, random_state=0) by default.
        :param n_jobs:  number of parallel jobs to use for calculation. if -1, all available cores are used. default 1.
        :param train_sizes: train sizes to tests
        :param path_save: path to save the plot. If None , lot is not saved. 'plot_learning_curve' by default.
        :param show: Whether to display the graphic. False by default.
        :param scalability: whether to retrieve scalability plots. Fit times per training examples.
        :param performance: whether to retrieve performance plots . Plot fit_time vs score
        :return: None
        https://scikit-learn.org/stable/auto_examples/model_selection/plot_learning_curve.html#sphx-glr-auto-examples-model-selection-plot-learning-curve-py
        """
        if classifier is None:
            classifier = self.classifier

        train_sizes, train_scores, test_scores, fit_times, _ = \
            learning_curve(classifier, self.x_train, self.y_train, cv=cv, n_jobs=n_jobs,
                           train_sizes=train_sizes,
                           return_times=True)
        train_scores_mean = np.mean(train_scores, axis=1)
        train_scores_std = np.std(train_scores, axis=1)
        test_scores_mean = np.mean(test_scores, axis=1)
        test_scores_std = np.std(test_scores, axis=1)
        fit_times_mean = np.mean(fit_times, axis=1)
        fit_times_std = np.std(fit_times, axis=1)

        # learning curve
        plt.clf()
        plt.grid(b=True, which='both')
        plt.figure()
        plt.title(title)
        if ylim is not None:
            plt.ylim(*ylim)
        plt.xlabel("Training examples")
        plt.ylabel("Score")
        plt.fill_between(train_sizes, train_scores_mean - train_scores_std,
                         train_scores_mean + train_scores_std, alpha=0.1,
                         color="r")
        plt.fill_between(train_sizes, test_scores_mean - test_scores_std,
                         test_scores_mean + test_scores_std, alpha=0.1, color="g")
        plt.plot(train_sizes, train_scores_mean, 'o-', color="r",
                 label="Training score")
        plt.plot(train_sizes, test_scores_mean, 'o-', color="g",
                 label="Cross-validation score")
        plt.legend(loc="best")
        if path_save is not None:
            plt.savefig(fname=str(path_save + '_learning_curve.png'))
        if show is True:
            plt.show()
        plt.clf()

        if scalability:
            # Plot n_samples vs fit_times
            plt.grid(b=True, which='both')
            plt.plot(train_sizes, fit_times_mean, 'o-')
            plt.fill_between(train_sizes, fit_times_mean - fit_times_std,
                             fit_times_mean + fit_times_std, alpha=0.1)
            plt.xlabel("Training examples")
            plt.ylabel("fit_times")
            plt.title("Scalability of the model")
            if path_save is not None:
                plt.savefig(fname=str(path_save + '_scalability_model.png'))
            if show is True:
                plt.show()
            plt.clf()

        if performance:
            # Plot fit_time vs score
            plt.grid(b=True, which='both')
            plt.plot(fit_times_mean, test_scores_mean, 'o-')
            plt.fill_between(fit_times_mean, test_scores_mean - test_scores_std,
                             test_scores_mean + test_scores_std, alpha=0.1)
            plt.xlabel("fit_times")
            plt.ylabel("Score")
            plt.title("Performance of the model")
            if path_save is not None:
                plt.savefig(fname=str(path_save + '_performance_model.png'))
            if show is True:
                plt.show()
            plt.clf()

        # check if it is better way to do this. option of the plots. arrange the model names. in ones it gets the model names and the others dont
        return

    ####################################################################################################################
    # FEATURE IMPORTANCE
    ####################################################################################################################

    # todo adicionar modelos
    def _get_feat_impo(self, classifier=None, column_to_sort='mean_coef'):
        if classifier is None:
            classifier = self.classifier

        if self.model_name in ('rf', 'gboosting', 'rfr', 'gboostingr'):
            feature_importance = pd.DataFrame(self.classifier.named_steps['clf'].feature_importances_,
                                              index=self.columns,
                                              columns=['importance']).sort_values('importance', ascending=False)
        elif self.model_name in ('svc', 'sgd', 'lr', 'linear_svc', 'tweedier','linear_svr', 'gpr', 'svr'):
            df = pd.DataFrame(self.classifier.named_steps['clf'].coef_.T,
                              index=self.columns)  # columns=['Coefficients'])

            # get a column that is the media of the modules of all classes
            df['mean_coef'] = abs(df).mean(axis=1)

            # sort the df by specific class, by avg of all classes
            feature_importance = df.iloc[(-df[column_to_sort].abs()).argsort()]
            # check if better this way or ordered positive negative

            # other hypothesis to get the avg of abs values per row
            # filter_col = [col for col in df if col.startswith('coef_abs')]
            # df['mean_coef_abs'] = df[filter_col].mean(axis=1)
            # get the module of each class coefficient
            # for column in df:
            #     column_name = 'coef_abs_' + str(column)
            #     df[column_name] = abs(df[column])

        else:
            e = "Model not supported, please choose between " \
                "'svc', 'sgd', 'linear_svc', 'lr', 'rf', 'rfr','gboosting'," \
                "'gboostingr', 'tweedier','linear_svr', 'gpr', 'svr' "
            print(e)
            return e
        return feature_importance

    def features_importances_df(self, classifier=None, model_name=None, top_features=20, column_to_sort='mean_coef'):
        """
        Function that given a classifier retrieves the features importances as a dataframe.
        :param classifier: {classifier instance} classifier or validation curve (e.g. sklearn.svm.SVC).If None (default)
        uses the classifier inside the class.
        :param model_name: model used in classifier. Choose between 'svm', 'sgd', 'lr', 'gboosting' or 'rf'.
        this only matters to determine _coef or _featureimportance
        :param top_features: number of features to display on the dataframe. 20 by default.
        :param column_to_sort: column to sort the dataframe. 'mean_coef' by default. This column represents the average
        scores in multiclass problems. coef for each class are also generated
        :return: table with features names and importance for the model
        """

        saved_args = locals()
        # if self.model_name not in ('rf', 'rfr', 'gboosting', 'svm', 'sgd', 'lr', 'linear_svm', 'tweedier_poisson'):
        #     print("Model not supported, please choose between 'svm', 'sgd', 'linear_svm', 'lr', 'rf', 'rfr', or 'gboosting'")
        #     return
        # do something for gboostingr? histgboostingr? that give something like features seen in
        if model_name is not None:
            self.model_name = model_name
        if classifier is not None:
            self.classifier = classifier

        feat_table = self._get_feat_impo(self.classifier, column_to_sort=column_to_sort)
        self.feat_table = feat_table

        if self.report_name is not None:
            self._report([self.features_importances_df.__name__, saved_args])
            self._report(feat_table[30:], dataframe=True, header=True, float_format='%.3f')
            with open(str(self.report_name[:-4] + 'fi.txt'), 'a+') as f:
                feat_table.to_csv(f, sep='\t', mode='a')
        return feat_table

    def features_importances_plot(self, classifier=None, top_features=20, model_name=None,
                                  column_to_plot=None,
                                  show=False, path_save='feat_impo.png',
                                  title=None,
                                  kind='barh', figsize=(9, 7), color='r', edgecolor='black', **params):
        """
        Function that given a classifier retrieves the features importances represented as barplot.
        :param classifier: {classifier instance} classifier or validation curve (e.g. sklearn.svm.SVC).If None (default)
        uses the classifier inside the class.
        :param model_name: model used in classifier. Choose between 'svm', 'sgd', 'lr', 'gboosting' or 'rf'
        :param top_features: number of features to display on plot. 20by default.
        :param column_to_plot: None by default. Can be given a specific class label. If None, will retrieve the
        average column of scores.  in case of models like svm, sgd, the coefs are retrieved from each class.
        if no class is given it gives a plot with average of absolute values and the top features considering that
        if a column is given, it re arranges and plot top features for that specific class.
        :param path_save: path to save the plot. If None , plot is not saved. 'feat_impo.png' by default.
        :param show: Whether to display the graphic. False by default.
        :param title: title of the plot. None by default. If  none title is 'Feature Importance Plot for {}' model name
        :param color: color of plot bard
        :param kind: bar parameter. horizontal bars per default 'barh'. other hypothesis include 'bar', 'box'
        :param figsize: size of the fig. (9,7) by default.
        :param edgecolor: 'black' by default
        :return:
        https://pandas.pydata.org/pandas-docs/version/0.23.4/generated/pandas.DataFrame.plot.html
        """

        saved_args = locals()
        if model_name is None:
            model_name = self.model_name
        else:
            self.model_name = model_name
        if classifier is None:
            classifier = self.classifier
        if self.feat_table is None:
            self.feat_table = self._get_feat_impo(classifier, column_to_plot)

        if title is None:
            title = 'Feature Importance Plot for {}'.format(model_name)

        color = color
        linewidth = 0.5
        label = 'Features Coefficients'

        if self.model_name.lower() in ('rf', 'gboosting', 'rfr', 'gboostingr'):
            label = 'Features Importance'
            table = self.feat_table[:top_features]

        # por isto no init ? o in in?
        elif self.model_name.lower() in ('svc', 'sgd', 'lr', 'linear_svc', 'svr', 'linear_svr', 'sgdr', 'tweedier'):
            label = 'Features Coefficients'
            if column_to_plot is None:  # all classes joined
                table = self.feat_table[:top_features]

                color_labels = np.unique(self.y_train)
                # List of colors in the color palettes
                rgb_values = sns.color_palette("nipy_spectral", len(color_labels) + 1)  # 'set2'
                color = rgb_values
                linewidth = 0.0

            else:  # graphic per class
                table = self.feat_table[column_to_plot][:top_features]
        else: # todo fix if it is other model rather the ones in the package por um try except?
            print("Model not supported, please choose between"
                  "'svc', 'sgd', 'linear_svc', 'lr', 'rf', 'rfr','gboosting',"
                  "'gboostingr', 'tweedier','linear_svr', 'gpr', 'svr'")
            return
        plt.clf()
        plt.title(title)
        plt.axvline(x=0, color='.5')
        plt.subplots_adjust(left=.3)
        table.plot(kind=kind, figsize=figsize, color=color, edgecolor=edgecolor, linewidth=linewidth, **params)

        plt.xlabel(label)
        plt.ylabel('Features')
        if path_save is not None:
            plt.savefig(fname=path_save)
        if show:
            plt.show()
        plt.clf()
        return

    ####################################################################################################################
    # PREDICTIONS
    ####################################################################################################################

    @timer
    def predict(self, x, seqs=None, classifier=None, names=None, true_y=None):

        """This function can be used to predict novel data with a trained classifier model. The function returns a
        'pandas.DataFrame' with predictions using the specified estimator and tests data.

        :param classifier: {classifier instance} classifier or validation curve (e.g. sklearn.svm.SVC).If None (default)
        uses the classifier inside the class.
        :param x: {array} descriptor values of the peptides to be predicted.
        :param seqs: {list} sequences of the peptides in ``x``. None by default
        :param names: {list} (optional) names of the peptides in ``x``. None by default.
        :param y: {array} (optional) true (known) classes of the peptides. None by default.
        :return: ``pandas.DataFrame`` containing predictions for ``x``. ``P_class0`` and ``P_class1``
            are the predicted probability of the peptide belonging to class 0 and class 1, respectively.

        Based on a function from moodlamp
        Müller A. T. et al. (2017) modlAMP: Python for anitmicrobial peptides, Bioinformatics 33, (17), 2753-2755,
        DOI:10.1093/bioinformatics/btx285
       """
        saved_args = locals()
        if classifier is None:
            classifier = self.classifier

        predict = classifier.predict(x)

        predict_df = pd.DataFrame(columns=['names', 'sequence', 'class predicted', 'True classes'])
        predict_df['names'] = names
        predict_df['True classes'] = true_y
        predict_df['sequence'] = seqs
        predict_df['class predicted'] = predict
        # remove columns with Nans all (cols that were None in entry
        predict_df = predict_df.dropna(axis=1)
        predict_df = predict_df.round(4)

        try:
            preds = classifier.predict_proba(x)
            cols_prob = ['prob_class_{}'.format(x) for x in range(self.final_units)]
            preds_df = pd.DataFrame(preds, columns=cols_prob)
            predict_df[cols_prob] = preds
            predict_df = predict_df.round(4)
        except Exception as e:
            print('classifier does not support predict probabilities')
            print(str(e))
            return predict_df

        if self.report_name is not None:
            self._report([self.predict.__name__, saved_args])
            predict_df.to_csv(str(self.report_name + 'predict' + time.strftime("-%Y%m%-%H%M%S.csv")))

        return predict_df

# todo PREDICT WINDOW needs to do also regression
# # alternatively the user can use the function GetSubSeq(self,protein_sequence,ToAA='S',window=3)
# # from the readsequence module
# # that Get all 2*window+1 sub-sequences whose center is ToAA (a specific aminoacid) in a protein
# # giving a ToAA and a window
#
# def add_features(self,list_of_sequences,list_functions):
#     """
#     Calculate all features available in package or lis of descriptors given
#     :param list_functions: list of features to be calculated with function adaptable from module Descriptors
#     :param list_of_sequences: list of sequences to calculate features
#     :return: dataframe with sequences and its features
#     """
#
#     features_list=[] #creating an empty list of dataset rows
#     for seq in list_of_sequences:
#         res={'sequence':seq}
#         sequence=ReadSequence() #creating sequence object
#         ps=sequence.read_protein_sequence(seq)
#         protein = Descriptor(ps) # creating object to calculate descriptors)
#         if len(list_functions)==0:
#             feature=protein.get_all(tricomp=True, bin_aa=False, bin_prop=False)
#         else:
#             feature=protein.adaptable(list_functions)
#         res.update(feature)
#         features_list.append(res)
#
#     df = pd.DataFrame(features_list)
#     #df.set_index(['sequence'],inplace=True)
#     return df
# #
# def predict_window(self,classifier, seq,x=None, window_size=20,gap=1,features=[], features_names=None,
#                    features_dataframe=None, names=None, y=None,filename=None):
#     """Scan a protein in a sliding window approach to predict novel peptides with a trained classifier model.
#     The function returns a
#     'pandas.DataFrame' with predictions using the specified estimator and tests data. If true class is provided,
#     it returns the scoring value for the tests data.
#     The user can provide a features_dataframe if does not want to use the descriptors supported. Otherwise, the user
#     can also provide a list with the numbers of the descriptors that want to be calculated (descriptors module adaptable())
#     if none is provided, the function will calculate all the descriptors available.
#
#     :param classifier: {classifier instance} classifier used for predictions.
#     :param x: {array} descriptor values of the peptides to be predicted.
#     :param seq:  sequence of the peptides in ``x``.
#     :param window_size: number of aminoacids to considerer in each seq to check. for default 20
#     :param gap: gap size of the search of windows in sequence. default 1
#     :param features: list containing the list features to be calculated under de adaptable function.
#             if list is empty will calculate all the descriptors available in the package
#     :param features_dataframe: dataframe with sequences and its features
#     :param features_names:names of features. If none will match the features with the ones with the model given
#     :param names: {list} (optional) names of the peptides in ``x``.
#     :param y: {array} (optional) true (known) classes of the peptides.
#     :param filename: {string} (optional) output filename to store the predictions to (``.csv`` format); if ``None``:
#         not saved.
#     :return: ``pandas.DataFrame`` containing predictions for subsequences generated, ``P_class0`` and ``P_class1``
#         are the predicted probability of the peptide belonging to class 0 and class 1, respectively. The previsions
#         are divided in probability classes <0.99, >0.95, >0.9, <0.8, >0.7, >0.6 and 0
#    """
#
#     # generate final list of sequences/split sequences
#     list_of_sequences,indices=sub_seq_sliding_window(seq,window_size,gap,index=True)
#
#
#     # calculate features for sequences
#
#     #if a dataframe with features calculated that will be considered
#     #if not features will be calculated.
#     # if features is none will calculate all the features according to the package and choose the features for the classifier
#     # if features is list with numbers will call the adaptable function
#
#     if features_dataframe != None: featuresDF=features_dataframe
#     else: featuresDF = self.add_features(list_of_sequences,features)
#
#     #select the features used for the model construction
#     if features_names==None:
#         features_to_select=self.X_data.columns
#     else:
#         features_to_select= features_names
#
#     x_predict_data=featuresDF[features_to_select]
#
#     # raw proability predictions of belonging or not
#     preds = classifier.predict_proba(x_predict_data)
#     #0 or 1 if belong or not
#     predict=classifier.predict(x_predict_data)
#
#     # dataframe with probabilities
#     df_pred= self.predict(classifier, x_predict_data, list_of_sequences, names=names, y=y, filename=filename)
#     df_pred['prevision']= predict
#     df_pred['pos_0']=[i[0] for i in indices]
#     df_pred['pos_-1']=[i[1] for i in indices]
#
#     # create a dataframe specifying really the sequences, indices, 1 and 0 and probability
#     # (can put scale >0.99, 095, 0.90, 0.8, 0.7, <0.7
#     column=['sequence', 'prevision','probability','scale_probability', 'pos_0', 'pos_-1']
#     sequence=''
#     pos_0=int
#     pos_1=int
#     rows = []
#
#     for index, row in df_pred.iterrows():
#         value=row['P_class1']
#         if value>=0.99:
#             pos_0=row['pos_0']
#             pos_1=row['pos_-1']
#             rows.append({'sequence': index, 'prevision': 1, 'probability': value, 'scale_probability':5,
#                          'pos_0':pos_0, 'pos_-1':pos_1 })
#
#         if value>=0.95 and value<0.99:
#             pos_0=row['pos_0']
#             pos_1=row['pos_-1']
#             rows.append({'sequence': index, 'prevision': 1, 'probability': value, 'scale_probability':4,
#                          'pos_0':pos_0, 'pos_-1':pos_1 })
#
#         if value>=0.90 and value<0.95:
#             pos_0=row['pos_0']
#             pos_1=row['pos_-1']
#             rows.append( {'sequence': index, 'prevision': 1, 'probability': value, 'scale_probability':3,
#                           'pos_0':pos_0, 'pos_-1':pos_1 })
#
#         if value>=0.80 and value<0.90:
#             pos_0=row['pos_0']
#             pos_1=row['pos_-1']
#             rows.append( {'sequence': index, 'prevision': 1, 'probability':value, 'scale_probability':2,
#                           'pos_0':pos_0, 'pos_-1':pos_1 })
#
#         if value>=0.70 and value<0.80:
#             pos_0=row['pos_0']
#             pos_1=row['pos_-1']
#             rows.append( {'sequence': index, 'prevision': 1, 'probability': value, 'scale_probability':1,
#                           'pos_0':pos_0, 'pos_-1':pos_1 })
#
#         if value>=0.60 and value<0.70:
#             pos_0=row['pos_0']
#             pos_1=row['pos_-1']
#             rows.append({'sequence': index, 'prevision': 1, 'probability': value, 'scale_probability':0,
#                          'pos_0':pos_0, 'pos_-1':pos_1 })
#
#         if value<0.6:
#             pos_0=row['pos_0']
#             pos_1=row['pos_-1']
#             rows.append({'sequence': index, 'prevision': 0, 'probability': value, 'scale_probability':0,
#                          'pos_0':pos_0, 'pos_-1':pos_1 })
#
#     # create a new dataframe where consecutive rows in the same probability scale are joined and positions updated
#     df = pd.DataFrame(rows) # dataframe with scale probabilities
#
#     df['key'] = (df['scale_probability'] != df['scale_probability'].shift(1)).astype(int).cumsum()
#     # add a sentinel column that tracks which group of consecutive data each row applies to
#
#     df_new=df.drop(columns=['sequence', 'prevision'])
#     x=df_new.values.tolist() #not in pandas dataframe
#     remove_list=[]
#
#     for row in range(len(x)-1):
#         pos_fin=x[row][3]
#         pos_ini=x[row][2]
#         prob=round(x[row][0],3)
#         x[row][0] = round(x[row][0], 4)
#         scale_prob=x[row][1]
#         key=x[row][4]
#
#         if x[row][4]==x[row+1][4]:
#             remove_list.append(row+1) #select rows to delete
#
#     # update the positions. is not made in the same loop because in more than two consecutive rows it will
#     # update a row that will be deleted
#     for index in reversed(remove_list):
#         x[index-1][0]=x[index][0]
#
#     for index in sorted(remove_list, reverse=True): # remove rows that were joined
#         del x[index]
#
#     # add sequence
#     for row in range(len(x)):
#         seqs=seq[int(x[row][2]):int(x[row][3])]
#         x[row].append(seqs)
#
#     final_df=pd.DataFrame(x, columns=['probability','scale_probability','pos_0','pos_-1','key','sequence'])
#     final_df=final_df.drop(columns=['key'])
#     final_df = final_df[['pos_0','pos_-1','probability','scale_probability','sequence']]
#     return final_df



import sys
sys.path.append("../")
import os
import torch
import torch.utils.data as data_utils
import pickle
from propythia.dna.calculate_features import calculate_and_normalize
from propythia.dna.read_sequence import ReadDNA
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import StandardScaler
from propythia.dna.encoding import DNAEncoder

from .utils import seed_everything

def data_splitting(fps_x, fps_y, batch_size, train_size, test_size, validation_size):
    """
    Split data into train, test and validation sets.
    """
    
    if(train_size + test_size + validation_size != 1):
        raise ValueError("The sum of train_size, test_size and validation_size must be 1.")
    
    x, x_test, y, y_test = train_test_split(
        fps_x, fps_y,
        test_size=test_size,
        train_size=train_size + validation_size,
        stratify=fps_y
    )
    x_train, x_cv, y_train, y_cv = train_test_split(
        x, y,
        test_size=validation_size/(1-test_size),
        train_size=1-(validation_size/(1-test_size)),
        stratify=y
    )
    
    train_data = data_utils.TensorDataset(
        torch.tensor(x_train, dtype=torch.float),
        torch.tensor(y_train, dtype=torch.long)
    )
    test_data = data_utils.TensorDataset(
        torch.tensor(x_test, dtype=torch.float),
        torch.tensor(y_test, dtype=torch.long)
    )
    valid_data = data_utils.TensorDataset(
        torch.tensor(x_cv, dtype=torch.float),
        torch.tensor(y_cv, dtype=torch.long)
    )

    trainloader = data_utils.DataLoader(
        train_data,
        shuffle=True,
        batch_size=batch_size
    )
    testloader = data_utils.DataLoader(
        test_data,
        shuffle=True,
        batch_size=batch_size
    )
    validloader = data_utils.DataLoader(
        valid_data,
        shuffle=True,
        batch_size=batch_size
    )
    
    return trainloader, testloader, validloader, x_train

def prepare_data(data_dir, mode, batch_size, k, train_size=0.6, test_size=0.2, validation_size=0.2):
    """
    Prepare data for training and testing.
    :param data_dir: str, the path to the data directory.
    :param mode: str, the mode to use. Must be either 'descriptor', 'one_hot', 'chemical' or 'kmer_one_hot'.
    :param batch_size: int, the batch size to use.
    :param k: int, value for the kmer one hot encoding.
    :param train_size: float, the proportion of the data to use for training.
    :param test_size: float, the proportion of the data to use for testing.
    :param validation_size: float, the proportion of the data to use for validation.
    :return: trainloader: torch.utils.data.DataLoader, the training data.
    :return: testloader: torch.utils.data.DataLoader, the testing data.
    :return: validloader: torch.utils.data.DataLoader, the validation data.
    :return: input_size: int, the size of the input.
    :return: sequence_length: int, the size of the length of sequence.
    """
    
    seed_everything()
    
    fps_x_file = data_dir + '/fps_x_descriptor.pkl' if mode == 'descriptor' else data_dir + '/fps_x.pkl'
    fps_y_file = data_dir + '/fps_y_descriptor.pkl' if mode == 'descriptor' else data_dir + '/fps_y.pkl'
    
    # check if fps_x_file and fps_y_file exist
    if not os.path.isfile(fps_x_file) or not os.path.isfile(fps_y_file):
        
        # read data from csv
        reader = ReadDNA()
        data = reader.read_csv(filename=data_dir + '/dataset.csv', with_labels=True)
        
        # calculate features
        if mode == 'descriptor':
            fps_x, fps_y = calculate_and_normalize(data)
        else:
            fps_x = data['sequence'].values
            fps_y = data['label'].values
        
        # save fps_x and fps_y to files
        with open(fps_x_file, 'wb') as f:
            pickle.dump(fps_x, f)
        with open(fps_y_file, 'wb') as f:
            pickle.dump(fps_y, f)
    else:
        # load fps_x and fps_y from files
        with open(fps_x_file, 'rb') as f:
            fps_x = pickle.load(f)
        with open(fps_y_file, 'rb') as f:
            fps_y = pickle.load(f)
    
    if(mode in ['one_hot', 'chemical', 'kmer_one_hot']):
        encoder = DNAEncoder(fps_x)
        possibilities = {
            'one_hot': encoder.one_hot_encode,
            'chemical': encoder.chemical_encode,
            'kmer_one_hot': encoder.kmer_one_hot_encode
        }
        fps_x = possibilities[mode]() if mode != 'kmer_one_hot' else possibilities[mode](k)
    elif(mode == 'descriptor'):
        scaler = StandardScaler().fit(fps_x)
        fps_x = scaler.transform(fps_x)
        fps_y = fps_y.to_numpy()
    else:
        raise ValueError("mode must be either 'one_hot', 'descriptor', 'chemical' or 'kmer_one_hot_encode'.")
    
    trainloader, testloader, validloader, x_train = data_splitting(fps_x, fps_y, batch_size, train_size, test_size, validation_size)
    
    return trainloader, testloader, validloader, x_train.shape[-1], x_train.shape[1]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
##############################################################################

File containing a class intend to facilitate parameter optimization. This class is not supposed to be ran outside
the Shallow learning and Deep learning classes.

The functions are based on the package scikit learn

Authors: Ana Marta Sequeira

Date: 12/2020

Email:

##############################################################################
"""
import numpy as np
import pandas as pd
from time import time
from sklearn.model_selection import train_test_split, cross_val_score, GridSearchCV, RandomizedSearchCV
from sklearn.metrics import make_scorer
from sklearn.metrics import matthews_corrcoef


# returns the best model fit but dont know if it uses the callbacks and stuff for deep learning models
# https://www.tensorflow.org/tensorboard/hyperparameter_tuning_with_hparams
class ParamOptimizer():
    """
    Class to optimize model parameters
    """

    def __init__(self, estimator, optType, paramDic, dataX, datay, cv=5, n_iter_search=15, n_jobs=1,
                 scoring=make_scorer(matthews_corrcoef), model_name=None, refit=True):

        self.optType = optType
        self.paramDic = paramDic
        self.dataX = dataX
        self.datay = datay
        self.model = estimator
        self.cv = cv
        self.n_iter_search = n_iter_search
        self.n_jobs = n_jobs
        self.score = scoring
        self.model_name = model_name
        self.final_units = len(np.unique(list(datay)))
        self.gs = None
        self.refit = refit

    # Utility function to report best scores
    def report_top_models(self, gs=None, n_top=3):
        """
        Utility function to report the top models
        :param gs:
        :param n_top:
        :return:
        """

        if gs is None:
            gs = self.gs
        print(gs)
        results = gs.cv_results_
        list_to_write = []
        for i in range(1, n_top + 1):
            candidates = np.flatnonzero(results['rank_test_score'] == i)
            for candidate in candidates:
                s1 = "Model with rank: {0}\n".format(i)
                s2 = "Mean validation score: {0:.3f} (std: {1:.3f})\n".format(
                    results['mean_test_score'][candidate],
                    results['std_test_score'][candidate])
                s3 = ("Parameters: {0:}\n".format(results['params'][candidate]))  # todo this values with 4 decimasl
                s4 = "\n"
                print(s1, s2, s3, s4)
                list_to_write.append([s1, s2, s3, s4])
        return list_to_write

    def report_all_models(self, gs=None):
        """
        Utility function to report all models tested
        :param gs:
        :return:
        """
        if gs is None:
            gs = self.gs
        print(self.score)
        print(self.cv)
        s1 = "Best score (scorer: %s) and parameters from a %d-fold cross validation:\n" % \
             (self.score, self.cv)
        s2 = "MCC score:\t%.3f\n" % gs.best_score_
        s3 = "Parameters:\t%s\n" % gs.best_params_
        print(s1, s2, s3)
        # results = gs.cv_results_
        # means = []
        # stds = []
        # for scorer in self.score:
        #     mean = {}
        #     std = {}
        #     mean[scorer] = results['mean_test_%s' % (scorer)]
        #     std[scorer] = results['std_test_%s' % (scorer)]
        #     means.append(mean)
        #     stds.append(std)

        means = gs.cv_results_['mean_test_score']
        stds = gs.cv_results_['std_test_score']
        params = gs.cv_results_['params']
        for mean, stdev, param in zip(means, stds, params):
            print("%f (%f) with: %r" % (mean, stdev, param))
        s4 = pd.DataFrame(params)
        s4['means'] = means
        s4['stds'] = stds
        print(s4)
        s4 = s4.sort_values(by=['means'], ascending=False)
        cols = ['means', 'stds'] + [col for col in s4 if col not in ('means', 'stds')]
        df = s4[cols]
        return s1, s2, s3, df

    def _randomized_search(self):
        # run randomized search
        start = time()
        gs = RandomizedSearchCV(estimator=self.model,
                                param_distributions=self.paramDic,
                                scoring=self.score,
                                cv=self.cv,
                                n_jobs=self.n_jobs,
                                n_iter=self.n_iter_search, refit=self.refit)
        gs = gs.fit(self.dataX, self.datay)
        print("RandomizedSearchCV took %.2f seconds for %d candidates"
              " parameter settings." % ((time() - start), self.n_iter_search))
        self.gs = gs
        return gs

    def _grid_search(self):
        # run grid search
        gs = GridSearchCV(estimator=self.model,
                          param_grid=self.paramDic,
                          scoring=self.score,
                          cv=self.cv,
                          n_jobs=self.n_jobs, refit=self.refit)
        start = time()
        gs.fit(self.dataX, self.datay)
        print("GridSearchCV took %.2f seconds for %d candidate parameter settings."
              % (time() - start, len(gs.cv_results_['params'])))

        return gs

    def get_opt_params(self):
        """
        Function to do gridSearch or randomizedSearch optimization
        :return: gridsearch object with metrics for the models and best model.  best model is accessed using
        gs.best_estimator_
        """
        if self.optType == 'gridSearch':
            gs = self._grid_search()
            return gs
        elif self.optType == 'randomizedSearch':
            gs = self._randomized_search()
            return gs
        else:
            print('Invalid optimization type!')
            return None



# -*- coding: utf-8 -*-
"""
#####################################################################################
Checking whether the input protein sequence is valid amino acid or not

Authors: Ana Marta Sequeira

Date:

Email:

#####################################################################################

"""
AALetter=["A","R","N","D","C","E","Q","G","H","I","L","K","M","F","P","S","T","W","Y","V"]

def protein_check(ProteinSequence):
	"""
	Check whether the protein sequence is a valid amino acid sequence or not

	:param ProteinSequence: protein sequence
	:return: length of protein if no problem, 0 if protein is not valid
	"""
	global flag
	num_pro=len(ProteinSequence)
	for i in ProteinSequence:
		if i not in AALetter:
			flag=0
			break
		else:
			flag=num_pro
	return flag

if __name__ == "__main__":

	protein_inv="ADGCGVGEGTGQGPMCNCMCMKWVYADEDAADLESDSFADEDASLESDSFPWSNQRVFCSFADEDASU"
	print(protein_check(protein_inv))
	protein='MQGNGSALPNASQPVLRGDGARPSWLASALACVLIFTIVVDILGNL'
	print(protein_check(protein))


# -*- coding: utf-8 -*-
"""
##############################################################################

File containing a class used for reducing the number of features on a dataset based
on unsupervised techniques. Feature decomposition.

Authors: Ana Marta Sequeira

Date: 06/2019 altered 12/2020

Email:

##############################################################################
"""
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
from matplotlib.lines import Line2D
from sklearn.decomposition import PCA, SparsePCA, MiniBatchSparsePCA, TruncatedSVD
from .ml.utils import timer

sns.set()


class FeatureDecomposition:
    """
    The Feature Reduction class aims to Reduce the number of features on a dataset based on unsupervised techniques.
    pca statistical procedure that orthogonally transforms the original n coordinates of a data set into a new set of
    n coordinates called principal components. Principal components are a combination of features that capture well the
    variance of the original features.
    Based on scikit learn.
    This class receives fps_x, a dataset containing the features or encodings. Classes of this representations may
    be given in the classes argument (optional). If wanted, a report name may be given and an automatically txt report
    will be generated with the results called with the class.
    """

    def __init__(self, fps_x, report_name=None, classes=None):
        """	constructor """
        self.fps_x = fps_x
        self.pca = None
        self.x_pca = None
        self.n_components = None
        self.labels = None  # to distinguish pca to sdv

        self.classes = classes  # columns names
        self.report_name = report_name
        if self.report_name:
            self.report(str(self.report_name))

    def report(self, info):
        filename = str(self.report_name)
        with open(filename, 'a+') as file:
            # print(info.type)
            if isinstance(info, str):
                file.writelines(info)
            else:
                for l in info:
                    file.writelines('\n{}'.format(l))

    @timer
    def run_pca(self, n_components=None, random_state=None, **params):
        """
        Function that realizes the pca analysis
        :param dataset: data on to perform pca
        :param n_components: Number of components to keep. if n_components is not set all components are kept. None
        by default.
        :param random_state: None by default.
        :return: pca and pca fitted a nd transformed to x objects.
        For more information: https://scikit-learn.org/stable/modules/generated/sklearn.decomposition.PCA.html
        """
        saved_args = locals()
        # performing pca
        pca_ = PCA(n_components, random_state, **params)
        x_pca_ = pca_.fit_transform(self.fps_x)

        # analysis
        s1 = "Original shape: {}".format(str(self.fps_x.shape))
        s2 = "Reduced shape: {}".format(str(x_pca_.shape))
        s3 = "Variance explained by the PC: {}".format(sum(pca_.explained_variance_ratio_))
        s4 = "Number of components: {}".format(pca_.n_components_)
        # s5 = "pca components by explained variance ratio:\n{}".format(pca_.components_)

        print('{}\n{}\n{}\n{}'.format(s1, s2, s3, s4))

        self.pca = pca_
        self.x_pca = x_pca_
        self.n_components = pca_.n_components_
        self.labels = 'PC-'

        if self.report_name:
            self.report([self.run_pca.__name__, saved_args, s1, s2, s3, s4, s5])

        return self.pca, self.x_pca
        # it is possible to access all the information in scikit learn

    @timer
    def run_batch_sparse_pca(self, n_components, alpha=1, batch_size=50, **params):
        """
        Function that realizes sparse pca analysis using batches.
        :param n_components: Number of components to keep. if n_components is not set all components are kept.
        :param alpha: alpha parameter. 1 by default.
        :param batch_size: batch size (number of samples to consider each time) to use when running batch PCA. 50 by default.
        :return: pca and pca fitted a nd transformed to x objects.
        For more information: https://scikit-learn.org/stable/modules/generated/sklearn.decomposition.PCA.html
        """
        # todo add explained_variance_ratio
        saved_args = locals()

        pca_ = MiniBatchSparsePCA(n_components, alpha=alpha, batch_size=batch_size, **params)
        x_pca_ = pca_.fit_transform(self.fps_x)

        # analysis
        s1 = "Original shape: {}".format(str(self.fps_x.shape))
        s2 = "Reduced shape: {}".format(str(x_pca_.shape))
        s4 = "Number of components: {}".format(pca_.n_components_)
        s5 = "pca components by explained variance ratio:\n{}".format(pca_.components_)

        print('{}\n{}\n{}'.format(s1, s2, s4))
        if self.report_name:
            self.report([self.run_batch_sparse_pca.__name__, saved_args, s1, s2, s4, s5])

        self.pca = pca_
        self.x_pca = x_pca_
        self.n_components = pca_.n_components_
        self.labels = 'spPC-'

        return self.pca, self.x_pca

    @timer
    def run_truncated_svd(self, n_components, n_iter=5, random_state=42, **params):
        """
        This transformer performs linear dimensionality reduction by means of truncated singular value decomposition(SVD).
        Contrary to PCA, this estimator does not center the data before computing the singular value
        decomposition. This means it can work with sparse matrices efficiently.
        https://machinelearningmastery.com/singular-value-decomposition-for-dimensionality-reduction-in-python/
         SVD is typically used on sparse data.
        :param n_components: Number of components to keep. if n_components is not set all components are kept. None
        by default.
        :param n_iter: number of iterations. 5 by default.
        :param random_state: None by default.
        :param params:other parameters for SVD function accepted for scikit learn
        :return: svd and svd fitted and transformed to x objects.
        """
        saved_args = locals()

        # performing svd
        svd = TruncatedSVD(n_components=n_components, n_iter=n_iter, random_state=random_state)
        x_svd = svd.fit_transform(self.fps_x)

        # analysis
        s1 = "Original shape: {}".format(str(self.fps_x.shape))
        s2 = "Reduced shape: {}".format(str(x_svd.shape))
        s3 = "Variance explained by the PC: {}".format(
            sum(svd.explained_variance_ratio_))  # Percentage of variance explained by each of the selected components
        s4 = "Number of components: {}".format(svd.components_)
        s5 = "pca components by explained variance ratio:\n{}".format(svd.components_)

        print('{}\n{}\n{}\n{}'.format(s1, s2, s3, s4))

        self.pca = svd
        self.x_pca = x_svd
        self.n_components = len(svd.components_)
        self.labels = 'SV'

        if self.report_name:
            self.report([self.run_truncated_svd.__name__, saved_args, s1, s2, s3, s4, s5])

        return self.pca, self.x_pca

    #####################################################################################################
    def variance_ratio_components(self):
        """
        measures the variance ratio of the principal components
        :return: variance ratio of principal components
        """

        ex_variance = np.var(self.x_pca, axis=0)
        ex_variance_ratio = ex_variance / np.sum(ex_variance)
        if self.report_name:
            self.report([self.variance_ratio_components.__name__, ex_variance_ratio])

        return ex_variance_ratio

    def contribution_of_features_to_component(self):
        """
        Function that retrieves a dataframe containing the contribution of each feature (rows) for component
        As unsupervised learning does not represent the importance of features but representing the directions
        of maximum variance in the data.
        :return: dataframe containing the contribution of each feature (rows) for component
        """
        # does not work with sparse
        # todopass this to the init
        original_columns = []
        if self.classes is not None:
            original_columns = self.classes
        else:
            try:
                original_columns = self.fps_x.columns
            except Exception as e:
                print(e)

        coef = self.pca.components_.T
        columns = []
        for x in range(self.n_components):
            columns.append(str(self.labels + str(x + 1)))
        result = pd.DataFrame(coef, columns=columns, index=original_columns)
        if self.report_name:
            self.report([self.contribution_of_features_to_component.__name__, result])
        return result

    def pca_bar_plot(self, show=True, path_save='pca_bar_plot.png',
                     title='Percentage of explained variance ratio by PCA',
                     width=1, data=None, color='b', edgecolor='k', linewidth=0,
                     tick_label=None, **params):
        """
        Function that derives a bar plot representing the percentage of explained variance ratio by pca.
        :param pca: dataset fit to pca
        :param height: scalar or sequence of scalars. The height(s) of the bars.
        :param width:  scalar or array-like, optional. The width(s) of the bars
        :param data:
        :param color: scalar or array-like, optional. The colors of the bar faces.
        :param edgecolor:  scalar or array-like, optional. The colors of the bar edges.
        :param linewidth: scalar or array-like, optional. Width of the bar edge(s). If 0, don't draw edges.
        :param tick_label: string or array-like, optional. The tick labels of the bars. Default: None
        :return: None
        For more information please see https://matplotlib.org/api/_as_gen/matplotlib.pyplot.bar.html
        """
        # does not work with sparse
        plt.clf()
        plt.bar(range(self.n_components), height=self.pca.explained_variance_ratio_ * 100, width=width, data=data,
                color=color,
                edgecolor=edgecolor, linewidth=linewidth, tick_label=tick_label, **params)
        plt.xlabel(str(self.labels))
        plt.ylabel('Percentage of explained variance')
        plt.title(title)
        if path_save is not None: plt.savefig(fname=path_save)
        if show is True: plt.show()
        plt.clf()

    def pca_cumulative_explain_ratio(self, title = 'cumulative explain ratio', show=True, path_save='pca_cumulative_explain_ratio.png'):
        """
        Plot cumulative explain variance per number of components. useful to understand how many components are needed
        to describe the data.
        :param show: Whether to show the plot or no. True by default.
        :param path_save: Path to save the plot. pca_cumulative_explain_ratio.png by default.
        :return: None
        """
        plt.clf()
        plt.plot(np.cumsum(self.pca.explained_variance_ratio_))
        plt.xlabel('number of components')
        plt.ylabel('cumulative explained variance')
        plt.title(title)
        if path_save is not None: plt.savefig(fname=path_save)
        if show is True: plt.show()
        plt.clf()

    def pca_scatter_plot(self, target, pca1=0, pca2=1, title=None, show=True, path_save='pca_scatter_plot.png'):
        """
        Scatter plot of the labels based on two components (by default the first ones).
        Retrieves graphic showing the positions of labels according of the two chosen components
        :param title: string. title of the scatter plot
        :param target: labels of dataset
        :param pca1: first pca to be considered. default PCA1
        :param pca2: second pca to be considered. default PCA2
        :param show: Whether to show the plot or no. True by default.
        :param path_save: Path to save the plot. pca_scatter_plot.png by default.
        :return: None
        """
        plt.clf()
        projected = self.x_pca

        n_color = len(np.unique(target))
        # Get Unique ec
        color_labels = np.unique(target)

        # List of colors in the color palettes
        rgb_values = sns.color_palette("nipy_spectral", n_color)  # 'set2'

        # Map ec to the colors
        color_map = dict(zip(color_labels, rgb_values))

        # Finally use the mapped values
        plt.scatter(projected[:, pca1], projected[:, pca2],
                    c=target.map(color_map))

        # create legend
        # classes  - colour labels
        # class_colours rgb values
        recs = []
        from matplotlib.lines import Line2D

        for i in range(0, n_color):
            recs.append(Line2D((0, 0.75), (0, 0), color=rgb_values[i], marker='o', linestyle=''))
            # recs.append(mpatches.Circle((0, 0), 0.5, fc=rgb_values[i]))

        n_col = int(n_color / 23) + 1

        lgd = plt.legend(recs, color_labels, bbox_to_anchor=(1, 1), loc="upper left",
                         fontsize='xx-small', ncol=n_col, shadow=False, title='classes')
        # add labels and title
        plt.xlabel(str(self.labels + '-1'))
        plt.ylabel(str(self.labels + '-2'))
        if title is None:
            title = ' Scatter plot of 2 {}'.format(str(self.labels))
        plt.title(title)

        # save and empty plt
        if path_save is not None:
            plt.savefig(fname=path_save, bbox_extra_artists=(lgd,), bbox_inches='tight', pad_inches=0)
        if show is True:
            plt.show()
        plt.clf()

        # plt.scatter(projected[:, pca1], projected[:, pca2],
        #             c=target,
        #             edgecolor='none', alpha=0.5)
        #             # ,
        # palette=sns.color_palette('nipy_spectral', n_color))
        # cmap=plt.cm.get_cmap('nipy_spectral',n_color))
        # plt.colorbar()
        # plt.legend(legend1,loc='best', shadow=False)

    def pca_scatter_plot3d(self, target, pca1=0, pca2=1, pca3=2, title=None, show=True,
                           path_save='pca_scatter_plot.png'):
        """
        Scatter plot of the labels based on three components (by default the first ones).
        Retrieves graphic showing the positions of labels according of the two chosen components
        :param title: string. title of the scatter plot
        :param target: labels of dataset
        :param pca1: first pca to be considered. default PCA1
        :param pca2: second pca to be considered. default PCA2
        :param pca3: third pca to be considered. default PCA3
        :param show: Whether to show the plot or no. True by default.
        :param path_save: Path to save the plot. pca_scatter_plot.png by default.
        :return: None
        """
        plt.clf()
        projected = self.x_pca

        n_color = len(np.unique(target))
        # Get Unique ec
        color_labels = np.unique(target)

        # List of colors in the color palettes
        rgb_values = sns.color_palette("nipy_spectral", n_color)  # 'set2'

        # Map ec to the colors
        color_map = dict(zip(color_labels, rgb_values))
        # Finally use the mapped values
        from mpl_toolkits import mplot3d
        ax = plt.axes(projection="3d")

        ax.scatter3D(projected[:, pca1], projected[:, pca2], projected[:, pca3],
                     c=target.map(color_map))
        # create legend
        # classes  - colour labels
        # class_colours rgb values
        recs = []

        for i in range(0, n_color):
            recs.append(Line2D((0, 0.75), (0, 0), color=rgb_values[i], marker='o', linestyle=''))
            # recs.append(mpatches.Circle((0, 0), 0.5, fc=rgb_values[i]))

        n_col = int(n_color / 23) + 1

        lgd = plt.legend(recs, color_labels, bbox_to_anchor=(1, 1), loc="upper left",
                         fontsize='xx-small', ncol=n_col, shadow=False, title='classes')
        # add labels and title
        ax.set_xlabel(str(self.labels + '-1'), fontweight='bold')
        ax.set_ylabel(str(self.labels + '-2'), fontweight='bold')
        ax.set_zlabel(str(self.labels + '-3'), fontweight='bold')
        if title is None:
            title = ' Scatter plot of 3 {}'.format(str(self.labels))
        plt.title(title)

        # save and empty plt
        if path_save is not None:
            plt.savefig(fname=path_save, bbox_extra_artists=(lgd,), bbox_inches='tight', pad_inches=0)
        if show is True:
            plt.show()
        plt.clf()




# -*- coding: utf-8 -*-
"""
#####################################################################################

This module is used for computing the composition, transition and distribution 

based on the different properties of AADs. The AADs with the same

properties is marked as the same number. You can get 147  for a given

protein sequence. You can freely use and distribute it. If you hava  any problem, 

you could contact with us timely!

References:

[1]: Inna Dubchak, Ilya Muchink, Stephen R.Holbrook and Sung-Hou Kim. Prediction 

of protein folding class using global description of amino acid sequence. Proc.Natl.

Acad.Sci.USA, 1995, 92, 8700-8704.

[2]:Inna Dubchak, Ilya Muchink, Christopher Mayor, Igor Dralyuk and Sung-Hou Kim. 

Recognition of a Protein Fold in the Context of the SCOP classification. Proteins: 

Structure, Function and Genetics,1999,35,401-407.

Authors: Dongsheng Cao and Yizeng Liang.

Date: 2010.11.22

Email: oriental-cds@163.com

Altered and converted to python 3.6 by Ana Marta Sequeira 05/2019

#####################################################################################

"""

import math, copy


AALetter=["A","R","N","D","C","E","Q","G","H","I","L","K","M","F","P","S","T","W","Y","V"]

_Hydrophobicity = {'1':'RKEDQN','2':'GASTPHY','3':'CLVIMFW'}
#'1'stand for Polar; '2'stand for Neutral, '3' stand for Hydrophobicity

_NormalizedVDWV = {'1':'GASTPD','2':'NVEQIL','3':'MHKFRYW'}
#'1'stand for (0-2.78); '2'stand for (2.95-4.0), '3' stand for (4.03-8.08)

_Polarity = {'1':'LIFWCMVY','2':'CPNVEQIL','3':'KMHFRYW'}
#'1'stand for (4.9-6.2); '2'stand for (8.0-9.2), '3' stand for (10.4-13.0)

_Charge = {'1':'KR','2':'ANCQGHILMFPSTWYV','3':'DE'}
#'1'stand for Positive; '2'stand for Neutral, '3' stand for Negative

_SecondaryStr = {'1':'EALMQKRH','2':'VIYCWFT','3':'GNPSD'}
#'1'stand for Helix; '2'stand for Strand, '3' stand for coil

_SolventAccessibility = {'1':'ALFCGIVW','2':'RKQEND','3':'MPSTHY'}
#'1'stand for Buried; '2'stand for Exposed, '3' stand for Intermediate

_Polarizability = {'1':'GASDT','2':'CPNVEQIL','3':'KMHFRYW'}
#'1'stand for (0-0.108); '2'stand for (0.128-0.186), '3' stand for (0.219-0.409)


##You can continuely add other properties of AADs to compute  of protein sequence.

_AATProperty=(_Hydrophobicity,_NormalizedVDWV,_Polarity,_Charge,_SecondaryStr,_SolventAccessibility,_Polarizability)

_AATPropertyName=('_Hydrophobicity','_NormalizedVDWV','_Polarity','_Charge','_SecondaryStr','_SolventAccessibility','_Polarizability')


##################################################################################################


def stringto_num(protein_sequence, aa_property):
	"""
	Tranform the protein sequence into the string form such as 32123223132121123.

	:param protein_sequence:  protein is a pure protein sequence.
	:param aa_property: AAProperty is a dict form containing classifciation of amino acids such as _Polarizability
	:return: result is a string such as 123321222132111123222
	"""

	hard_protein_sequence=copy.deepcopy(protein_sequence)
	for k,m in list(aa_property.items()):
		for index in m:
			hard_protein_sequence=str.replace(hard_protein_sequence,index,k)
	t_protein_sequence=hard_protein_sequence

	return t_protein_sequence


def calculate_composition(protein_sequence, aa_property, aap_name):
	"""
	A method used for computing composition

	:param protein_sequence: protein is a pure protein sequence
	:param aa_property: dict form containing classifciation of amino acids such as _Polarizability.
	:param aap_name: string used for indicating a AAP name.
	:return: dict form containing composition based on the given property.
	"""
	t_protein_sequence=stringto_num(protein_sequence, aa_property)
	result={}
	num=len(t_protein_sequence)
	result[aap_name + 'C' + '1']=round(float(t_protein_sequence.count('1')) / num, 3)
	result[aap_name + 'C' + '2']=round(float(t_protein_sequence.count('2')) / num, 3)
	result[aap_name + 'C' + '3']=round(float(t_protein_sequence.count('3')) / num, 3)
	return result


def calculate_transition(protein_sequence, aa_property, aap_name):
	"""
	A method used for computing transition

	:param protein_sequence: protein is a pure protein sequence
	:param aa_property: AAProperty is a dict form containing classifciation of amino acids such as _Polarizability
	:param aap_name: string used for indicating a AAP name
	:return: dict form containing transition based on the given property
	"""
	t_protein_sequence=stringto_num(protein_sequence, aa_property)
	result={}
	num=len(t_protein_sequence)
	CTD=t_protein_sequence
	result[aap_name + 'T' + '12']=round(float(CTD.count('12') + CTD.count('21')) / (num - 1), 3)
	result[aap_name + 'T' + '13']=round(float(CTD.count('13') + CTD.count('31')) / (num - 1), 3)
	result[aap_name + 'T' + '23']=round(float(CTD.count('23') + CTD.count('32')) / (num - 1), 3)
	return result


def calculate_distribution(protein_sequence, aa_property, aap_name):
	"""
	A method used for computing distribution

	:param protein_sequence: protein is a pure protein sequence
	:param aa_property: dict form containing classification of amino acids such as _Polarizability
	:param aap_name: string used for indicating a AAP name
	:return: dict form containing Distribution based on the given property
	"""
	t_protein_sequence=stringto_num(protein_sequence, aa_property)
	result={}
	num=len(t_protein_sequence)
	temp=('1','2','3')
	for i in temp:
		num=t_protein_sequence.count(i)
		ink=1
		indexk=0
		cds=[]
		while ink<=num:
			indexk=str.find(t_protein_sequence,i,indexk)+1
			cds.append(indexk)
			ink=ink+1
				
		if cds==[]:
			result[aap_name + 'D' + i + '001']=0
			result[aap_name + 'D' + i + '025']=0
			result[aap_name + 'D' + i + '050']=0
			result[aap_name + 'D' + i + '075']=0
			result[aap_name + 'D' + i + '100']=0
		else:
				
			result[aap_name + 'D' + i + '001']=round(float(cds[0]) / num * 100, 3)
			result[aap_name + 'D' + i + '025']=round(float(cds[int(math.floor(num * 0.25)) - 1]) / num * 100, 3)
			result[aap_name + 'D' + i + '050']=round(float(cds[int(math.floor(num * 0.5)) - 1]) / num * 100, 3)
			result[aap_name + 'D' + i + '075']=round(float(cds[int(math.floor(num * 0.75)) - 1]) / num * 100, 3)
			result[aap_name + 'D' + i + '100']=round(float(cds[-1]) / num * 100, 3)

	return result

##################################################################################################


def calculate_composition_hydrophobicity(protein_sequence):
	"""
	A method used for calculating composition based on Hydrophobicity of AADs.

	:param protein_sequence: protein is a pure protein sequence.
	:return: dict form containing Composition based on Hydrophobicity.
	"""
	result=calculate_composition(protein_sequence, _Hydrophobicity, '_Hydrophobicity')
	return result


def calculate_composition_normalized_vdwv(protein_sequence):
	"""
	A method used for calculating composition based on NormalizedVDWV of AADs.

	:param protein_sequence: protein is a pure protein sequence
	:return: dict form containing Composition based on NormalizedVDWV
	"""
	result=calculate_composition(protein_sequence, _NormalizedVDWV, '_NormalizedVDWV')
	return result


def calculate_composition_polarity(protein_sequence):
	"""
	calculating composition based on Polarity of AADs.

	:param protein_sequence: protein is a pure protein sequence.
	:return: dict form containing Composition based on Polarity
	"""
	
	result=calculate_composition(protein_sequence, _Polarity, '_Polarity')
	return result


def calculate_composition_charge(protein_sequence):
	"""
	alculating composition based on Charge of AADs.

	:param protein_sequence: protein is a pure protein sequence.
	:return: dict form containing Composition based on Charge
	"""
	result=calculate_composition(protein_sequence, _Charge, '_Charge')
	return result


def calculate_composition_secondary_str(protein_sequence):
	"""
	 method used for calculating composition based on SecondaryStr of AADs

	:param protein_sequence: protein is a pure protein sequence.
	:return: dict form containing Composition based on SecondaryStr
	"""
	result=calculate_composition(protein_sequence, _SecondaryStr, '_SecondaryStr')
	return result


def calculate_composition_solvent_accessibility(protein_sequence):
	"""
	A method used for calculating composition based on SolventAccessibility of  AADs.

	:param protein_sequence: protein is a pure protein sequence.
	:return: dict form containing Composition based on SolventAccessibility
	"""
	result=calculate_composition(protein_sequence, _SolventAccessibility, '_SolventAccessibility')
	return result


def calculate_composition_polarizability(protein_sequence):
	"""
	A method used for calculating composition based on Polarizability of AADs.

	:param protein_sequence: protein is a pure protein sequence.
	:return: ict form containing Composition based on Polarizability
	"""
	
	result=calculate_composition(protein_sequence, _Polarizability, '_Polarizability')
	return result

##################################################################################################


def calculate_transition_hydrophobicity(protein_sequence):
	"""
	A method used for calculating Transition based on Hydrophobicity of AADs.

	:param protein_sequence: protein is a pure protein sequence
	:return: dict form containing Transition based on Hydrophobicity.
	"""
	result=calculate_transition(protein_sequence, _Hydrophobicity, '_Hydrophobicity')
	return result


def calculate_transition_normalized_vdwv(protein_sequence):
	"""
	A method used for calculating Transition based on NormalizedVDWV of AADs.

	:param protein_sequence: protein is a pure protein sequence.
	:return: result is a dict form containing Transition based on NormalizedVDWV.
	"""
	result=calculate_transition(protein_sequence, _NormalizedVDWV, '_NormalizedVDWV')
	return result


def calculate_transition_polarity(protein_sequence):
	"""
	A method used for calculating Transition based on Polarity of	AADs.

	:param protein_sequence: protein is a pure protein sequence.
	:return: result is a dict form containing Transition based on Polarity.
	"""
	result=calculate_transition(protein_sequence, _Polarity, '_Polarity')
	return result


def calculate_transition_charge(protein_sequence):
	"""
	A method used for calculating Transition based on Charge of AADs.

	:param protein_sequence: protein is a pure protein sequence.
	:return: result is a dict form containing Transition based on Charge.
	"""
	result=calculate_transition(protein_sequence, _Charge, '_Charge')
	return result


def calculate_transition_secondary_str(protein_sequence):
	"""
	A method used for calculating Transition based on SecondaryStr of	AADs

	:param protein_sequence: protein is a pure protein sequence
	:return: result is a dict form containing Transition based on SecondaryStr
	"""
	result=calculate_transition(protein_sequence, _SecondaryStr, '_SecondaryStr')
	return result


def calculate_transition_solvent_accessibility(protein_sequence):
	"""
	A method used for calculating Transition based on SolventAccessibility of  AADs.

	:param protein_sequence: protein is a pure protein sequence
	:return: dict form containing Transition based on SolventAccessibility.
	"""
	result=calculate_transition(protein_sequence, _SolventAccessibility, '_SolventAccessibility')
	return result


def calculate_transition_polarizability(protein_sequence):
	"""
	A method used for calculating Transition based on Polarizability of AADs.

	:param protein_sequence: protein is a pure protein sequence
	:return: dict form containing Transition based on Polarizability.
	"""
	result=calculate_transition(protein_sequence, _Polarizability, '_Polarizability')
	return result

##################################################################################################


def calculate_distribution_hydrophobicity(protein_sequence):
	"""
	A method used for calculating Distribution based on Hydrophobicity of AADs.

	:param protein_sequence: protein is a pure protein sequence
	:return: dict form containing Distribution based on Hydrophobicity.
	"""
	result=calculate_distribution(protein_sequence, _Hydrophobicity, '_Hydrophobicity')
	return result


def calculate_distribution_normalized_vdwv(protein_sequence):
	"""
	A method used for calculating Distribution based on NormalizedVDWV of AADs

	:param protein_sequence: protein is a pure protein sequence.
	:return: dict form containing Distribution based on NormalizedVDWV
	"""
	result=calculate_distribution(protein_sequence, _NormalizedVDWV, '_NormalizedVDWV')
	return result


def calculate_distribution_polarity(protein_sequence):
	"""
	A method used for calculating Distribution based on Polarity of AADs

	:param protein_sequence: protein is a pure protein sequence.
	:return: dict form containing Distribution based on Polarity
	"""
	result=calculate_distribution(protein_sequence, _Polarity, '_Polarity')
	return result


def calculate_distribution_charge(protein_sequence):
	"""
	A method used for calculating Distribution based on Charge of	AADs.

	:param protein_sequence: protein is a pure protein sequence.
	:return: dict form containing Distribution based on Charge
	"""
	result=calculate_distribution(protein_sequence, _Charge, '_Charge')
	return result
	
def calculate_distribution_secondary_str(protein_sequence):
	"""
	A method used for calculating Distribution based on SecondaryStr of AADs.

	:param protein_sequence: protein is a pure protein sequence.
	:return: dict form containing Distribution based on SecondaryStr.
	"""
	result=calculate_distribution(protein_sequence, _SecondaryStr, '_SecondaryStr')
	return result
	
def calculate_distribution_solvent_accessibility(protein_sequence):
	"""
	A method used for calculating Distribution based on SolventAccessibility of  AADs.

	:param protein_sequence: protein is a pure protein sequence
	:return: dict form containing Distribution based on SolventAccessibility
	"""
	result=calculate_distribution(protein_sequence, _SolventAccessibility, '_SolventAccessibility')
	return result
	
def calculate_distribution_polarizability(protein_sequence):
	"""
	A method used for calculating Distribution based on Polarizability of	AADs.

	:param protein_sequence: protein is a pure protein sequence.
	:return: dict form containing Distribution based on Polarizability
	"""
	result=calculate_distribution(protein_sequence, _Polarizability, '_Polarizability')
	return result

##################################################################################################


def calculate_c(protein_sequence):
	"""
	Calculate all composition seven different properties of AADs.

	:param protein_sequence: protein is a pure protein sequence.
	:return: dict form containing all composition
	"""
	result={}
	result.update(calculate_composition_polarizability(protein_sequence))
	result.update(calculate_composition_solvent_accessibility(protein_sequence))
	result.update(calculate_composition_secondary_str(protein_sequence))
	result.update(calculate_composition_charge(protein_sequence))
	result.update(calculate_composition_polarity(protein_sequence))
	result.update(calculate_composition_normalized_vdwv(protein_sequence))
	result.update(calculate_composition_hydrophobicity(protein_sequence))
	return result


def calculate_t(protein_sequence):
	"""
	Calculate all transition based seven different properties of AADs.

	:param protein_sequence: protein is a pure protein sequence
	:return: dict form containing all transition
	"""
	result={}
	result.update(calculate_transition_polarizability(protein_sequence))
	result.update(calculate_transition_solvent_accessibility(protein_sequence))
	result.update(calculate_transition_secondary_str(protein_sequence))
	result.update(calculate_transition_charge(protein_sequence))
	result.update(calculate_transition_polarity(protein_sequence))
	result.update(calculate_transition_normalized_vdwv(protein_sequence))
	result.update(calculate_transition_hydrophobicity(protein_sequence))
	return result
	

def calculate_d(protein_sequence):
	"""
	Calculate all distribution based seven different properties of AADs.

	:param protein_sequence: protein is a pure protein sequence
	:return: dict form containing all distribution
	"""
	result={}
	result.update(calculate_distribution_polarizability(protein_sequence))
	result.update(calculate_distribution_solvent_accessibility(protein_sequence))
	result.update(calculate_distribution_secondary_str(protein_sequence))
	result.update(calculate_distribution_charge(protein_sequence))
	result.update(calculate_distribution_polarity(protein_sequence))
	result.update(calculate_distribution_normalized_vdwv(protein_sequence))
	result.update(calculate_distribution_hydrophobicity(protein_sequence))
	return result


def calculate_ctd(protein_sequence):
	"""
	Calculate all CTD based seven different properties of AADs

	:param protein_sequence: protein is a pure protein sequence.
	:return: result is a dict form containing all C
	"""
	result={}
	result.update(calculate_composition_polarizability(protein_sequence))
	result.update(calculate_composition_solvent_accessibility(protein_sequence))
	result.update(calculate_composition_secondary_str(protein_sequence))
	result.update(calculate_composition_charge(protein_sequence))
	result.update(calculate_composition_polarity(protein_sequence))
	result.update(calculate_composition_normalized_vdwv(protein_sequence))
	result.update(calculate_composition_hydrophobicity(protein_sequence))
	result.update(calculate_transition_polarizability(protein_sequence))
	result.update(calculate_transition_solvent_accessibility(protein_sequence))
	result.update(calculate_transition_secondary_str(protein_sequence))
	result.update(calculate_transition_charge(protein_sequence))
	result.update(calculate_transition_polarity(protein_sequence))
	result.update(calculate_transition_normalized_vdwv(protein_sequence))
	result.update(calculate_transition_hydrophobicity(protein_sequence))
	result.update(calculate_distribution_polarizability(protein_sequence))
	result.update(calculate_distribution_solvent_accessibility(protein_sequence))
	result.update(calculate_distribution_secondary_str(protein_sequence))
	result.update(calculate_distribution_charge(protein_sequence))
	result.update(calculate_distribution_polarity(protein_sequence))
	result.update(calculate_distribution_normalized_vdwv(protein_sequence))
	result.update(calculate_distribution_hydrophobicity(protein_sequence))
	return result


if __name__ == "__main__":
	
#	import scipy,string

#	result=scipy.zeros((268,147))
#	f=file('protein1.txt','r')
#	for i,j in enumerate(f:
#		temp=calculate_ctd(string.strip(j))
#		result[i,:]=temp.values()
#	scipy.savetxt('ResultNCTRER.txt', result, fmt='%15.5f',delimiter='')
#
	protein="ADGCGVGEGTGQGPMCNCMCMKWVYADEDAADLESDSFADEDASLESDSFPWSNQRVFCSFADEDAS"
#	print stringto_num(protein,_Hydrophobicity)
#	print calculate_composition(protein,_Hydrophobicity,'_Hydrophobicity')
#	print calculate_transition(protein,_Hydrophobicity,'_Hydrophobicity')
#	print calculate_distribution(protein,_Hydrophobicity,'_Hydrophobicity')
#	print calculate_distribution_solvent_accessibility(protein)
#	print len(calculate_ctd(protein))
#	print len(calculate_c(protein))
#	print len(calculate_t(protein))
#	print len(calculate_d(protein))
	print(calculate_ctd(protein))





# -*- coding: utf-8 -*-
"""

###############################################################################

The module is used for computing the composition of amino acids, dipetide and 

3-mers (tri-peptide) for a given protein sequence. You can get 8420 descriptors

for a given protein sequence. You can freely use and distribute it. If you hava 

any problem, you could contact with us timely!

References:

[1]: Reczko, M. and Bohr, H. (1994) The DEF data base of sequence based protein

fold class predictions. Nucleic Acids Res, 22, 3616-3619.

[2]: Hua, S. and Sun, Z. (2001) Support vector machine approach for protein

subcellular localization prediction. Bioinformatics, 17, 721-728.


[3]:Grassmann, J., Reczko, M., Suhai, S. and Edler, L. (1999) Protein fold class

prediction: new methods of statistical classification. Proc Int Conf Intell Syst Mol

Biol, 106-112.

Authors: Dongsheng Cao and Yizeng Liang.

Date: 2012.3.27

Email: oriental-cds@163.com

Altered and converted to python 3.6 for Ana Marta Sequeira 05/2019

###############################################################################
"""

import re

AALetter=["A","R","N","D","C","E","Q","G","H","I","L","K","M","F","P","S","T","W","Y","V"]
#############################################################################################


def calculate_aa_composition(protein_sequence):
	"""
	Calculate the composition of Amino acids for a given protein sequence.

	:param protein_sequence: protein is a pure protein sequence.
	:return: result is a dict form containing the composition of 20 amino acids.
	"""

	length_sequence=len(protein_sequence)
	result={}
	for i in AALetter:
		result[i] = round(float(protein_sequence.count(i)) / length_sequence * 100, 3)
	return result
#############################################################################################

def calculate_dipeptide_composition(protein_sequence):
	"""
	Calculate the composition of dipeptidefor a given protein sequence.

	:param protein_sequence: protein is a pure protein sequence.
	:return: result is a dict form containing the composition of 400 dipeptides.
	"""
	length_sequence=len(protein_sequence)
	result={}
	for i in AALetter:
		for j in AALetter:
			dipeptide=i+j
			result[dipeptide]=round(float(protein_sequence.count(dipeptide)) / (length_sequence - 1) * 100, 2)
	return result
#############################################################################################


def getkmers():
	"""
	Get the amino acid list of 3-mers.

	:return: result is a list form containing 8000 tri-peptides.
	"""
	kmers = list()
	for i in AALetter:
		for j in AALetter:
			for k in AALetter:
				kmers.append(i+j+k)
	return kmers

#############################################################################################


def get_spectrum_dict(protein_sequence):
	"""
	Calculate the spectrum of 3-mers for a given protein.

	:param protein_sequence: protein is a pure protein sequence.
	:return: result is a dict form containing the composition values of 8000
	"""

	result = {}
	kmers = getkmers()
	for i in kmers:
		result[i]=len(re.findall(i, protein_sequence))
	return result

#############################################################################################


def calculate_aa_tripeptide_composition(protein_sequence):
	"""
	Calculate the composition of AADs, dipeptide and 3-mers for a given protein sequence.

	:param protein_sequence: protein is a pure protein sequence.
	:return: result is a dict form containing all composition values of AADs, dipeptide and 3-mers (8420).
	"""
	result={}
	result.update(calculate_aa_composition(protein_sequence))
	result.update(calculate_dipeptide_composition(protein_sequence))
	result.update(get_spectrum_dict(protein_sequence))

	return result


if __name__ == "__main__":

	protein="ADGCGVGEGTGQGPMCNCMCMKWVYADEDAADLESDSFADEDASLESDSFPWSNQRVFCSFADEDAS"

	AAC=calculate_aa_composition(protein)
	print(AAC)
	DIP=calculate_dipeptide_composition(protein)
	print(DIP)
	spectrum=get_spectrum_dict(protein)
	print(spectrum)
	res=calculate_aa_tripeptide_composition(protein)
	print(len(res))





"""
#####################################################################################
Preprocessing the amino acid sequences.

Authors: Miguel Barros

Date: 02/2022

Email:

#####################################################################################

"""
AALetter=["A","R","N","D","C","E","Q","G","H","I","L","K","M","F","P","S","T","W","Y","V"]


def protein_preprocessing(ProteinSequence, B :str ='N', Z : str = 'Q', U :str = 'C', O: str = 'K', J : str = 'I', X :str = ''):
	'''
	Edits a protein sequence by replacing aminoacids like Asparagine (B),  Glutamine(Z), Selenocysteine (U) and
         Pyrrolysine (O), by default they are replaced for the closest aminoacid residue if is present in the sequence, Asparagine (N), Glutamine (Q),
         Cysteinen(C) and Lysine (K), respectively. It also removes the ambiguous aminoacid (X) and alters the ambiguous
         aminoacid (J), by default it is replaced for Isoleucine (I). Use with caution.

	:param ProteinSequence: Protein sequence
	:param B: One-letter aminoacid code to replace Asparagine (B). Default amino acid is Asparagine (N)
	:param Z: One-letter aminoacid code to replace Glutamine(Z). Default amino acid is Glutamine (Q)
	:param U: One-letter aminoacid code to replace Selenocysteine (U). Default amino acid is Cysteinen(C)
	:param O: One-letter aminoacid code to replace Pyrrolysine (O). Default amino acid is Lysine (K)
	:param J: One-letter aminoacid code to replace ambiguous aminoacid (J). Default amino acid is Isoleucine (I)
	:param X: One-letter aminoacid code to replace ambiguous aminoacid (X), by default it is removed.
	:return: transformed protein sequence
	'''
	Seq = ProteinSequence.replace('B', B).replace('Z', Z).replace('U',U).replace('O', O).replace('X', X)\
		.replace('J', J)
	return Seq

